# -*- coding: utf-8 -*-

import pytest
from typing import Tuple

from gordo import _parse_version, __version__


def test_version():
    assert isinstance(__version__, str)
    versions: (int, int, int) = _parse_version(__version__)
    for v in versions:
        assert isinstance(v, int)


@pytest.mark.parametrize(
    "version,expected",
    [
        ("1.1.1", (1, 1)),
        ("1.1.1.dev-a1", (1, 1)),
        ("0.55.25.02", (0, 55)),
        ("0.0.0", (0, 0)),
    ],
)
def test_version_parser(version: str, expected: Tuple[int, int, int]):
    assert _parse_version(version) == expected



# -*- coding: utf-8 -*-
import logging
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from dateutil import tz
from typing import Iterable, List, Optional

import pandas as pd
from azure.datalake.store import core

from gordo.machine.dataset.data_provider.azure_utils import walk_azure
from gordo.machine.dataset.data_provider.base import GordoBaseDataProvider
from gordo.machine.dataset.sensor_tag import SensorTag
from gordo.machine.dataset.sensor_tag import to_list_of_strings
from gordo.util import capture_args

logger = logging.getLogger(__name__)


class IrocReader(GordoBaseDataProvider):
    ASSET_TO_PATH = {
        "ninenine": "/raw/plant/uon/cygnet/ninenine/history",
        "uon_ef": "/raw/plant/uon/cygnet/efd/history",
    }

    def can_handle_tag(self, tag: SensorTag):
        return IrocReader.base_path_from_asset(tag.asset) is not None

    @capture_args
    def __init__(self, client: core.AzureDLFileSystem, threads: int = 50, **kwargs):
        """
        Creates a reader for tags from IROC.
        """
        self.client = client
        self.threads = threads
        if self.threads is None:
            self.threads = 50
        logger.info(f"Starting IROC reader with {self.threads} threads")

    def load_series(
        self,
        train_start_date: datetime,
        train_end_date: datetime,
        tag_list: List[SensorTag],
        dry_run: Optional[bool] = False,
    ):
        """
        See GordoBaseDataProvider for documentation
        """
        if dry_run:
            raise NotImplementedError("Dry run for IrocReader is not implemented")
        if not tag_list:
            logger.warning("Iroc reader called with empty tag_list, returning none")
            return
        if train_end_date < train_start_date:
            raise ValueError(
                f"Iroc reader called with train_end_date: {train_end_date} before train_start_date: {train_start_date}"
            )

        base_paths_from_assets = list(
            map(lambda tag: IrocReader.base_path_from_asset(tag.asset), tag_list)
        )
        if len(set(base_paths_from_assets)) != 1:
            raise ValueError(
                "Iroc reader found either more than one asset or no asset from the tag list provided"
            )
        elif None in base_paths_from_assets:
            raise ValueError("Iroc reader could not associate some tags to an asset.")

        base_path = base_paths_from_assets[0]

        # We query with an extra day on both sides since the way the files are
        # organized in the datalake does not account for timezones, so some timestamps
        # are in the wrong files

        all_base_paths = (
            f"{base_path}/{t.year:0>4d}/{t.month:0>2d}/{t.day:0>2d}/"
            for t in pd.date_range(
                start=train_start_date.astimezone(tz.tzutc()) - pd.Timedelta("1D"),
                end=train_end_date.astimezone(tz.tzutc()) + pd.Timedelta("1D"),
                freq="D",
            )
        )

        fetched_tags = self._fetch_all_iroc_files_from_paths(
            all_base_paths, train_start_date, train_end_date, tag_list
        )
        if len(fetched_tags) < 0:
            raise ValueError(
                f"Found no data for tags {tag_list} in the daterange {train_start_date} to "
                f"{train_end_date}"
            )

        concatted = pd.concat(fetched_tags, copy=False)

        if len(concatted.columns) != len(tag_list):
            raise ValueError(
                f"Did not find data for all tags, the missing tags are "
                f"{set(to_list_of_strings(tag_list))-set(concatted.columns)}"
            )

        for col in concatted.columns:
            withouth_na = concatted[col].dropna()
            withouth_na.sort_index(inplace=True)
            yield withouth_na

    def _fetch_all_iroc_files_from_paths(
        self,
        all_base_paths: Iterable[str],
        train_start_date: datetime,
        train_end_date: datetime,
        tag_list: List[SensorTag],
    ):
        # Generator over all files in all of the base_paths
        def _all_files():
            for b_path in all_base_paths:
                for f in walk_azure(client=self.client, base_path=b_path):
                    yield f

        with ThreadPoolExecutor(max_workers=self.threads) as executor:
            # Pandas.concat makes the generator into a list anyway, so no extra memory
            # is used here
            fetched_tags = list(
                executor.map(
                    lambda file_path: self._read_iroc_df_from_azure(
                        file_path=file_path,
                        train_start_date=train_start_date,
                        train_end_date=train_end_date,
                        tag_list=tag_list,
                    ),
                    _all_files(),
                )
            )
            fetched_tags = [tags for tags in fetched_tags if not tags is None]

        return fetched_tags

    def _read_iroc_df_from_azure(
        self, file_path, train_start_date: datetime, train_end_date: datetime, tag_list
    ):
        adls_file_system_client = self.client

        logger.info("Attempting to open IROC file {}".format(file_path))

        try:
            with adls_file_system_client.open(file_path, "rb") as f:
                logger.info("Parsing file {}".format(file_path))
                df = read_iroc_file(f, train_start_date, train_end_date, tag_list)
            return df
        except:
            logger.warning(f"Problem parsing file {file_path}, skipping.")
            return None

    @staticmethod
    def base_path_from_asset(asset: str):
        """
        Resolves an asset code to the datalake basepath containing the data.
        Returns None if it does not match any of the asset codes we know.
        """
        if not asset:
            return None

        logger.debug(f"Looking for match for asset {asset}")
        asset = asset.lower()
        if asset not in IrocReader.ASSET_TO_PATH:
            logger.warning(
                f"Could not find match for asset {asset} in the list of "
                f"supported assets: {IrocReader.ASSET_TO_PATH.keys()}"
            )
            return None

        logger.debug(
            f"Found asset code {asset}, returning {IrocReader.ASSET_TO_PATH[asset]}"
        )
        return IrocReader.ASSET_TO_PATH[asset]


def read_iroc_file(
    file_obj,
    train_start_date: datetime,
    train_end_date: datetime,
    tag_list: List[SensorTag],
) -> pd.DataFrame:
    """
    Reads a single iroc timeseries csv, and returns it as a pandas.DataFrame.
    The returned dataframe has timestamps as a DateTimeIndex, and upto one column
    per tag in tag_list, but excluding any tags which does not exist in the csv.

    Parameters
    ----------
    file_obj: str or path object or file-like object
        File object to read iroc timeseries data from
    train_start_date
        Only keep timestamps later or equal than this
    train_end_date
        Only keep timestamps earlier than this
    tag_list
        Only keep tags in this list.

    Returns
    -------
    pd.DataFrame
        Dataframe with timestamps as a DateTimeIndex, and up to one column
        per tag in tag_list, but excluding any tags which does not exist in the
        csv.

    """
    df = pd.read_csv(file_obj, sep=",", usecols=["tag", "value", "timestamp"])
    df = df[df["tag"].isin(to_list_of_strings(tag_list))]
    # Note, there are some "digital" sensors with string values,
    # now they are just NaN converted
    df["value"] = df["value"].apply(pd.to_numeric, errors="coerce", downcast="float")
    df.dropna(inplace=True, subset=["value"])
    df["timestamp"] = pd.to_datetime(df["timestamp"], utc=True)
    df = df.pivot(index="timestamp", columns="tag")
    df = df[(df.index >= train_start_date) & (df.index < train_end_date)]
    df.columns = df.columns.droplevel(0)
    return df



import dictdiffer


def patch_dict(original_dict: dict, patch_dictionary: dict) -> dict:
    """Patches a dict with another. Patching means that any path defines in the
    patch is either added (if it does not exist), or replaces the existing value (if
    it exists). Nothing is removed from the original dict, only added/replaced.

    Parameters
    ----------
    original_dict : dict
        Base dictionary which will get paths added/changed
    patch_dictionary: dict
        Dictionary which will be overlaid on top of original_dict

    Examples
    --------
    >>> patch_dict({"highKey":{"lowkey1":1, "lowkey2":2}}, {"highKey":{"lowkey1":10}})
    {'highKey': {'lowkey1': 10, 'lowkey2': 2}}
    >>> patch_dict({"highKey":{"lowkey1":1, "lowkey2":2}}, {"highKey":{"lowkey3":3}})
    {'highKey': {'lowkey1': 1, 'lowkey2': 2, 'lowkey3': 3}}
    >>> patch_dict({"highKey":{"lowkey1":1, "lowkey2":2}}, {"highKey2":4})
    {'highKey': {'lowkey1': 1, 'lowkey2': 2}, 'highKey2': 4}

    Returns
    -------
    dict
        A new dictionary which is the result of overlaying `patch_dictionary` on top of
        `original_dict`

    """
    diff = dictdiffer.diff(original_dict, patch_dictionary)
    adds_and_mods = [(f, d, s) for (f, d, s) in diff if f != "remove"]
    return dictdiffer.patch(adds_and_mods, original_dict)



# -*- coding: utf-8 -*-
import os
import random
import logging
import threading
import timeit

from datetime import datetime
import typing

from cachetools import cached, TTLCache
import numpy as np
import pandas as pd
from influxdb import DataFrameClient

from gordo.machine.dataset.data_provider.azure_utils import create_adls_client
from gordo.machine.dataset.data_provider.base import GordoBaseDataProvider
from gordo.util import capture_args

from gordo.machine.dataset.data_provider.iroc_reader import IrocReader
from gordo.machine.dataset.data_provider.ncs_reader import NcsReader
from gordo.machine.dataset.sensor_tag import SensorTag


logger = logging.getLogger(__name__)


class NoSuitableDataProviderError(ValueError):
    pass


def load_series_from_multiple_providers(
    data_providers: typing.List[GordoBaseDataProvider],
    train_start_date: datetime,
    train_end_date: datetime,
    tag_list: typing.List[SensorTag],
    dry_run: typing.Optional[bool] = False,
) -> typing.Iterable[pd.DataFrame]:
    """
    Loads the tags in `tag_list` using multiple instances of
    :class:`gordo.machine.dataset.data_provider.base.GordoBaseDataProvider` provided in the
    parameter `data_providers`. Will load a tag from the first data provider in the list
    which claims it. See
    :func:`gordo.machine.dataset.data_provider.base.GordoBaseDataProvider.load_series`.

    Returns
    -------
    typing.Iterable[pd.Series]
        The required tags as an iterable of series where each series contains
        the tag values along with a datetime index.

    """
    readers_to_tags = {
        reader: [] for reader in data_providers
    }  # type: typing.Dict[GordoBaseDataProvider, typing.List[SensorTag]]

    for tag in tag_list:
        for tag_reader in data_providers:
            if tag_reader.can_handle_tag(tag):
                readers_to_tags[tag_reader].append(tag)
                logger.debug(f"Assigning tag: {tag} to reader {tag_reader}")
                # In case of a tag matching two readers, we let the "first"
                # one handle it
                break
        # The else branch is executed if the break is not called
        else:
            raise NoSuitableDataProviderError(
                f"Found no data providers able to download the tag {tag} "
            )
    before_downloading = timeit.default_timer()
    for tag_reader, readers_tags in readers_to_tags.items():
        if readers_tags:
            logger.debug(f"Using tag reader {tag_reader} to fetch tags {readers_tags}")
            for series in tag_reader.load_series(
                train_start_date=train_start_date,
                train_end_date=train_end_date,
                tag_list=readers_tags,
                dry_run=dry_run,
            ):
                yield series
    logger.debug(
        f"Downloading all tags took {timeit.default_timer()-before_downloading} seconds"
    )


class DataLakeProvider(GordoBaseDataProvider):

    _SUB_READER_CLASSES = [
        NcsReader,
        IrocReader,
    ]  # type: typing.List[typing.Type[GordoBaseDataProvider]]

    def can_handle_tag(self, tag):
        """Implements base method, see GordoBaseDataProvider"""
        for r in self._get_sub_dataproviders():
            if r.can_handle_tag(tag):
                return True
        return False

    @capture_args
    def __init__(
        self,
        storename: str = "dataplatformdlsprod",
        interactive: bool = False,
        dl_service_auth_str: str = None,
        **kwargs,
    ):
        """
        Instantiates a DataLakeBackedDataset, for fetching of data from the data lake

        Parameters
        ----------
        storename
            The store name to read data from
        interactive: bool
            To perform authentication interactively, or attempt to do it a
            automatically, in such a case must provide 'del_service_authS_tr'
            parameter or as 'DL_SERVICE_AUTH_STR' env var.
        dl_service_auth_str: Optional[str]
            string on the format 'tenant_id:service_id:service_secret'. To
            perform authentication automatically; will default to
            DL_SERVICE_AUTH_STR env var or None

        """
        self.storename = storename
        self.interactive = interactive
        self.dl_service_auth_str = dl_service_auth_str or os.environ.get(
            "DL_SERVICE_AUTH_STR"
        )
        self.client = None
        self.kwargs = kwargs
        self.lock = threading.Lock()

    def load_series(
        self,
        train_start_date: datetime,
        train_end_date: datetime,
        tag_list: typing.List[SensorTag],
        dry_run: typing.Optional[bool] = False,
    ) -> typing.Iterable[pd.Series]:
        """
        See
        :func:`gordo.machine.dataset.data_provider.base.GordoBaseDataProvider.load_series`
        for documentation
        """
        # We create them here so we only try to get a auth-token once we actually need
        # it, otherwise we would have constructed them in the constructor.
        if train_end_date < train_start_date:
            raise ValueError(
                f"DataLakeReader called with train_end_date: {train_end_date} before train_start_date: {train_start_date}"
            )
        data_providers = self._get_sub_dataproviders()

        yield from load_series_from_multiple_providers(
            data_providers, train_start_date, train_end_date, tag_list, dry_run
        )

    def _get_client(self):
        logger.debug("Acquiring threading lock for Datalake authentication.")
        with self.lock:
            if not self.client:
                self.client = create_adls_client(
                    storename=self.storename,
                    dl_service_auth_str=self.dl_service_auth_str,
                    interactive=self.interactive,
                )
        logger.debug("Released threading lock for Datalake authentication.")

        return self.client

    def _get_sub_dataproviders(self):
        data_providers = [
            t_reader(client=self._get_client(), **self.kwargs)
            for t_reader in DataLakeProvider._SUB_READER_CLASSES
        ]
        return data_providers


class InfluxDataProvider(GordoBaseDataProvider):
    @capture_args
    def __init__(
        self,
        measurement: str,
        value_name: str = "Value",
        api_key: str = None,
        api_key_header: str = None,
        client: DataFrameClient = None,
        uri: str = None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        measurement: str
            Name of the measurement to select from in Influx
        value_name: str
            Name of value to select, default to 'Value'
        api_key: str
            Api key to use in header
        api_key_header: str
            Key of header to insert the api key for requests
        uri: str
            Create a client from a URI
            format: <username>:<password>@<host>:<port>/<optional-path>/<db_name>
        kwargs: dict
            These are passed directly to the init args of influxdb.DataFrameClient
        """
        self.measurement = measurement
        self.value_name = value_name
        self.influx_client = client
        if kwargs.pop("threads", None):
            logger.warning(
                "InfluxDataProvider got parameter 'threads' which is not supported, it "
                "will be ignored."
            )

        if self.influx_client is None:
            if uri:

                # Import here to avoid any circular import error caused by
                # importing TimeSeriesDataset, which imports this provider
                # which would have imported Client via traversal of the __init__
                # which would then try to import TimeSeriesDataset again.
                from gordo.client.utils import influx_client_from_uri

                self.influx_client = influx_client_from_uri(  # type: ignore
                    uri,
                    api_key=api_key,
                    api_key_header=api_key_header,
                    dataframe_client=True,
                )
            else:
                if "type" in kwargs:
                    kwargs.pop("type")
                self.influx_client = DataFrameClient(**kwargs)
                if api_key is not None:
                    if not api_key_header:
                        raise ValueError(
                            "If supplying an api key, you must supply the header key to insert it under."
                        )
                    self.influx_client._headers[api_key_header] = api_key

    def load_series(
        self,
        train_start_date: datetime,
        train_end_date: datetime,
        tag_list: typing.List[SensorTag],
        dry_run: typing.Optional[bool] = False,
    ) -> typing.Iterable[pd.Series]:
        """
        See GordoBaseDataProvider for documentation
        """
        if dry_run:
            raise NotImplementedError(
                "Dry run for InfluxDataProvider is not implemented"
            )
        return (
            self.read_single_sensor(
                train_start_date=train_start_date,
                train_end_date=train_end_date,
                tag=tag.name,
                measurement=self.measurement,
            )
            for tag in tag_list
        )

    def read_single_sensor(
        self,
        train_start_date: datetime,
        train_end_date: datetime,
        tag: str,
        measurement: str,
    ) -> pd.Series:
        """
        Parameters
        ----------
            train_start_date: datetime
                Datetime to start querying for data
            train_end_date: datetime
                Datetime to stop query for data
            tag: str
                Name of the tag to match in influx
            measurement: str
                name of the measurement to select from
        Returns
        -------
            One column DataFrame
        """

        logger.info(f"Reading tag: {tag}")
        logger.info(f"Fetching data from {train_start_date} to {train_end_date}")
        query_string = f"""
            SELECT "{self.value_name}" as "{tag}"
            FROM "{measurement}"
            WHERE("tag" =~ /^{tag}$/)
                {f"AND time >= {int(train_start_date.timestamp())}s" if train_start_date else ""}
                {f"AND time <= {int(train_end_date.timestamp())}s" if train_end_date else ""}
        """

        logger.info(f"Query string: {query_string}")
        dataframes = self.influx_client.query(query_string)  # type: ignore

        try:
            df = list(dataframes.values())[0]
            return df[tag]

        except IndexError as e:
            list_of_tags = self._list_of_tags_from_influx()
            if tag not in list_of_tags:
                raise ValueError(f"tag {tag} is not found in influx")
            logger.error(
                f"Unable to find data for tag {tag} in the time range {train_start_date} - {train_end_date}"
            )
            raise e

    def _list_of_tags_from_influx(self):
        query_tags = (
            f"""SHOW TAG VALUES ON {self.influx_client._database} WITH KEY="tag" """
        )
        result = self.influx_client.query(query_tags)
        list_of_tags = []
        for item in list(result.get_points()):
            list_of_tags.append(item["value"])
        return list_of_tags

    @cached(cache=TTLCache(maxsize=10, ttl=600))
    def get_list_of_tags(self) -> typing.List[str]:
        """
        Queries Influx for the list of tags, using a TTL cache of 600 seconds. The
        cache can be cleared with :func:`cache_clear()` as is usual with cachetools.

        Returns
        -------
        typing.List[str]
            The list of tags in Influx

        """
        return self._list_of_tags_from_influx()

    def can_handle_tag(self, tag: SensorTag):
        return tag.name in self.get_list_of_tags()


class RandomDataProvider(GordoBaseDataProvider):
    """
    Get a GordoBaseDataset which returns unstructed values for X and y. Each instance
    uses the same seed, so should be a function (same input -> same output)
    """

    def can_handle_tag(self, tag: SensorTag):
        return True  # We can be random about everything

    @capture_args
    def __init__(self, min_size=100, max_size=300, **kwargs):
        self.max_size = max_size
        self.min_size = min_size
        np.random.seed(0)

    # Thanks stackoverflow
    # https://stackoverflow.com/questions/50559078/generating-random-dates-within-a-given-range-in-pandas
    @staticmethod
    def _random_dates(start, end, n=10):
        start = pd.to_datetime(start)
        end = pd.to_datetime(end)
        start_u = start.value // 10 ** 9
        end_u = end.value // 10 ** 9

        return sorted(
            pd.to_datetime(np.random.randint(start_u, end_u, n), unit="s", utc=True)
        )

    def load_series(
        self,
        train_start_date: datetime,
        train_end_date: datetime,
        tag_list: typing.List[SensorTag],
        dry_run: typing.Optional[bool] = False,
    ) -> typing.Iterable[pd.Series]:
        if dry_run:
            raise NotImplementedError(
                "Dry run for RandomDataProvider is not implemented"
            )
        for tag in tag_list:
            nr = random.randint(self.min_size, self.max_size)

            random_index = self._random_dates(train_start_date, train_end_date, n=nr)
            series = pd.Series(
                index=random_index,
                name=tag.name,
                data=np.random.random(size=len(random_index)),
            )
            yield series



# -*- coding: utf-8 -*-
import dateutil.parser
import os
import yaml
from typing import List, Optional, Dict, Any

import pytest
import numpy as np
import pandas as pd
import sklearn.compose
import sklearn.ensemble
from sklearn.base import BaseEstimator
from sklearn.multioutput import MultiOutputRegressor
from sklearn.linear_model import LinearRegression
from sklearn import metrics
from mock import patch

import gordo
from gordo.builder import ModelBuilder
from gordo.machine.dataset.sensor_tag import SensorTag
from gordo.machine.model import models
from gordo.machine import Machine
from gordo.machine.metadata import Metadata


def get_random_data():
    data = {
        "type": "RandomDataset",
        "train_start_date": dateutil.parser.isoparse("2017-12-25 06:00:00Z"),
        "train_end_date": dateutil.parser.isoparse("2017-12-30 06:00:00Z"),
        "tag_list": [SensorTag("Tag 1", None), SensorTag("Tag 2", None)],
        "target_tag_list": [SensorTag("Tag 1", None), SensorTag("Tag 2", None)],
    }
    return data


def machine_check(machine: Machine, check_history):
    """Helper to verify model builder metadata creation"""
    assert isinstance(machine.metadata.build_metadata.model.model_offset, int)

    # Scores is allowed to be an empty dict. in a case where the pipeline/transformer
    # doesn't implement a .score()
    if machine.metadata.build_metadata.model.cross_validation.scores != dict():
        tag_list = [tag.name.replace(" ", "-") for tag in machine.dataset.tag_list]
        scores_list = [
            "r2-score",
            "explained-variance-score",
            "mean-squared-error",
            "mean-absolute-error",
        ]
        all_scores_list = [
            f"{score}-{tag}" for score in scores_list for tag in tag_list
        ] + scores_list
        scores_metadata = machine.metadata.build_metadata.model.cross_validation.scores

        assert all(score in scores_metadata for score in all_scores_list)

    if check_history:
        assert "history" in machine.metadata.build_metadata.model.model_meta
        assert all(
            name in machine.metadata.build_metadata.model.model_meta["history"]
            for name in ("params", "loss", "accuracy")
        )


@pytest.mark.parametrize("scaler", [None, "sklearn.preprocessing.MinMaxScaler"])
def test_get_metrics_dict_scaler(scaler, mock):
    mock_model = mock
    metrics_list = [sklearn.metrics.mean_squared_error]
    # make the features in y be in different scales
    y = pd.DataFrame(
        np.array([[1, 1], [2, 2], [3, 3], [4, 4], [5, 5]]) * [1, 100],
        columns=["Tag 1", "Tag 2"],
    )
    metrics_dict = ModelBuilder.build_metrics_dict(metrics_list, y, scaler=scaler)
    metric_func = metrics_dict["mean-squared-error"]

    mock_model.predict = lambda _y: _y * [0.8, 1]
    mse_feature_one_wrong = metric_func(mock_model, y, y)
    mock_model.predict = lambda _y: _y * [1, 0.8]
    mse_feature_two_wrong = metric_func(mock_model, y, y)

    if scaler:
        assert np.isclose(mse_feature_one_wrong, mse_feature_two_wrong)
    else:
        assert not np.isclose(mse_feature_one_wrong, mse_feature_two_wrong)


@pytest.mark.parametrize(
    "model,expected_offset",
    (
        (models.KerasLSTMAutoEncoder(kind="lstm_hourglass", lookback_window=10), 9),
        (models.KerasLSTMForecast(kind="lstm_symmetric", lookback_window=13), 13),
        (models.KerasAutoEncoder(kind="feedforward_hourglass"), 0),
        (MultiOutputRegressor(LinearRegression()), 0),
    ),
)
def test_determine_offset(model: BaseEstimator, expected_offset: int):
    """
    Determine the correct output difference from the model
    """
    X, y = np.random.random((100, 10)), np.random.random((100, 10))
    model.fit(X, y)
    offset = ModelBuilder._determine_offset(model, X)
    assert offset == expected_offset


def test_output_dir(tmpdir):
    """
    Test building of model will create subdirectories for model saving if needed.
    """
    model_config = {"sklearn.decomposition.pca.PCA": {"svd_solver": "auto"}}
    data_config = get_random_data()
    output_dir = os.path.join(tmpdir, "some", "sub", "directories")
    machine = Machine(
        name="model-name", dataset=data_config, model=model_config, project_name="test"
    )
    builder = ModelBuilder(machine)
    model, machine_out = builder.build()
    machine_check(machine_out, False)

    builder._save_model(model=model, machine=machine_out, output_dir=output_dir)

    # Assert the model was saved at the location
    # Should be model file, and the metadata
    assert len(os.listdir(output_dir)) == 2


@pytest.mark.parametrize(
    "raw_model_config",
    (
        # Without pipeline
        """
    sklearn.preprocessing.MinMaxScaler:
        feature_range: [-1, 1]
    """,
        # Saves history
        """
    gordo.machine.model.models.KerasAutoEncoder:
        kind: feedforward_hourglass
    """,
        # With typical pipeline
        """
    sklearn.pipeline.Pipeline:
        steps:
          - sklearn.preprocessing.MinMaxScaler
          - sklearn.decomposition.pca.PCA:
              svd_solver: auto
    """,
        # Nested pipelilnes
        """
        sklearn.pipeline.Pipeline:
            steps:
              - sklearn.pipeline.Pipeline:
                  steps:
                    - sklearn.preprocessing.MinMaxScaler
              - sklearn.pipeline.Pipeline:
                  steps:
                    - sklearn.decomposition.pca.PCA:
                        svd_solver: auto
        """,
        # Pipeline as a parameter to another estimator
        """
        sklearn.compose.TransformedTargetRegressor:
            regressor:
                sklearn.pipeline.Pipeline:
                    steps:
                    - sklearn.preprocessing.MinMaxScaler
                    - gordo.machine.model.models.KerasAutoEncoder:
                        kind: feedforward_hourglass
                        compression_factor: 0.5
                        encoding_layers: 2
                        func: tanh
                        out_func: linear
                        epochs: 3
            transformer: sklearn.preprocessing.MinMaxScaler
    """,
    ),
)
def test_builder_metadata(raw_model_config):
    """
    Ensure the builder works with various model configs and that each has
    expected/valid metadata results.
    """
    model_config = yaml.load(raw_model_config, Loader=yaml.FullLoader)
    data_config = get_random_data()
    machine = Machine(
        name="model-name", dataset=data_config, model=model_config, project_name="test"
    )
    model, machine_out = ModelBuilder(machine).build()
    # Check metadata, and only verify 'history' if it's a *Keras* type model
    machine_check(machine_out, "Keras" in raw_model_config)


@pytest.mark.parametrize(
    "model,expect_empty_dict",
    (
        # Model is not a GordoBase, first parameter is a pipeline and no GordoBase either
        (
            sklearn.compose.TransformedTargetRegressor(
                regressor=sklearn.pipeline.Pipeline(
                    steps=[
                        ("s1", sklearn.preprocessing.MinMaxScaler()),
                        ("s2", sklearn.ensemble.RandomForestRegressor()),
                    ]
                ),
                transformer=sklearn.preprocessing.MinMaxScaler(),
            ),
            True,
        ),
        # Model is not a GordoBase, first parameter is not GordoBase either.
        (
            sklearn.compose.TransformedTargetRegressor(
                regressor=sklearn.ensemble.RandomForestRegressor(),
                transformer=sklearn.preprocessing.MinMaxScaler(),
            ),
            True,
        ),
        # Model is not a GordoBase, first parameter is a pipeline with GordoBase
        (
            sklearn.compose.TransformedTargetRegressor(
                regressor=sklearn.pipeline.Pipeline(
                    steps=[
                        ("s1", sklearn.preprocessing.MinMaxScaler()),
                        (
                            "s2",
                            gordo.machine.model.models.KerasAutoEncoder(
                                kind="feedforward_hourglass"
                            ),
                        ),
                    ]
                ),
                transformer=sklearn.preprocessing.MinMaxScaler(),
            ),
            False,
        ),
        # Model is not GordoBase but the first parameter is.
        (
            sklearn.compose.TransformedTargetRegressor(
                regressor=gordo.machine.model.models.KerasAutoEncoder(
                    kind="feedforward_hourglass"
                ),
                transformer=sklearn.preprocessing.MinMaxScaler(),
            ),
            False,
        ),
        # Plain model, no GordoBase
        (sklearn.ensemble.RandomForestRegressor(), True),
        # GordoBase bare
        (
            gordo.machine.model.models.KerasAutoEncoder(kind="feedforward_hourglass"),
            False,
        ),
    ),
)
def test_get_metadata_helper(model: BaseEstimator, expect_empty_dict: bool):
    """
    Ensure the builder works with various model configs and that each has
    expected/valid metadata results.
    """

    X, y = np.random.random((1000, 4)), np.random.random((1000,))

    model.fit(X, y)

    metadata = ModelBuilder._extract_metadata_from_model(model)

    # All the metadata we've implemented so far is 'history', so we'll check that
    if not expect_empty_dict:
        assert "history" in metadata
        assert all(
            name in metadata["history"] for name in ("params", "loss", "accuracy")
        )
    else:
        assert dict() == metadata


@pytest.mark.parametrize(
    "raw_model_config",
    (
        f"""
    gordo.machine.model.anomaly.diff.DiffBasedAnomalyDetector:
        scaler: sklearn.preprocessing.MinMaxScaler
        base_estimator:
            sklearn.compose.TransformedTargetRegressor:
                transformer: sklearn.preprocessing.MinMaxScaler
                regressor:
                    sklearn.pipeline.Pipeline:
                        steps:
                        - sklearn.preprocessing.MinMaxScaler
                        - gordo.machine.model.models.KerasAutoEncoder:
                            kind: feedforward_hourglass
                            batch_size: 3
                            compression_factor: 0.5
                            encoding_layers: 1
                            func: tanh
                            out_func: linear
                            epochs: 1
        """,
        f"""
   sklearn.compose.TransformedTargetRegressor:
       transformer: sklearn.preprocessing.MinMaxScaler
       regressor:
          sklearn.pipeline.Pipeline:
              steps:
              - sklearn.preprocessing.MinMaxScaler
              - gordo.machine.model.models.KerasAutoEncoder:
                  kind: feedforward_hourglass
                  batch_size: 2
                  compression_factor: 0.5
                  encoding_layers: 1
                  func: tanh
                  out_func: linear
                  epochs: 1
                 """,
        f"""
  sklearn.pipeline.Pipeline:
      steps:
      - sklearn.preprocessing.MinMaxScaler
      - gordo.machine.model.models.KerasAutoEncoder:
          kind: feedforward_hourglass
          batch_size: 2
          compression_factor: 0.5
          encoding_layers: 1
          func: tanh
          out_func: linear
          epochs: 1
         """,
    ),
)
def test_scores_metadata(raw_model_config):
    data_config = get_random_data()
    model_config = yaml.load(raw_model_config, Loader=yaml.FullLoader)
    machine = Machine(
        dataset=data_config, model=model_config, name="model-name", project_name="test"
    )
    model, machine_out = ModelBuilder(machine).build()
    machine_check(machine_out, False)


def test_output_scores_metadata():
    data_config = get_random_data()
    raw_model_config = f"""
            gordo.machine.model.anomaly.diff.DiffBasedAnomalyDetector:
                scaler: sklearn.preprocessing.MinMaxScaler
                base_estimator:
                    sklearn.compose.TransformedTargetRegressor:
                        transformer: sklearn.preprocessing.MinMaxScaler
                        regressor:
                            sklearn.pipeline.Pipeline:
                                steps:
                                - sklearn.preprocessing.MinMaxScaler
                                - gordo.machine.model.models.KerasAutoEncoder:
                                    kind: feedforward_hourglass
                                    batch_size: 3
                                    compression_factor: 0.5
                                    encoding_layers: 1
                                    func: tanh
                                    out_func: linear
                                    epochs: 1
            """

    model_config = yaml.load(raw_model_config, Loader=yaml.FullLoader)
    machine = Machine(
        name="model-name", dataset=data_config, model=model_config, project_name="test"
    )
    model, machine_out = ModelBuilder(machine).build()
    scores_metadata = machine_out.metadata.build_metadata.model.cross_validation.scores
    assert (
        scores_metadata["explained-variance-score-Tag-1"]["fold-mean"]
        + scores_metadata["explained-variance-score-Tag-2"]["fold-mean"]
    ) / 2 == pytest.approx(scores_metadata["explained-variance-score"]["fold-mean"])

    assert (
        scores_metadata["r2-score-Tag-1"]["fold-mean"]
        + scores_metadata["r2-score-Tag-2"]["fold-mean"]
    ) / 2 == pytest.approx(scores_metadata["r2-score"]["fold-mean"])

    assert (
        scores_metadata["mean-squared-error-Tag-1"]["fold-mean"]
        + scores_metadata["mean-squared-error-Tag-2"]["fold-mean"]
    ) / 2 == pytest.approx(scores_metadata["mean-squared-error"]["fold-mean"])

    assert (
        scores_metadata["mean-absolute-error-Tag-1"]["fold-mean"]
        + scores_metadata["mean-absolute-error-Tag-2"]["fold-mean"]
    ) / 2 == pytest.approx(scores_metadata["mean-absolute-error"]["fold-mean"])


def test_provide_saved_model_simple_happy_path(tmpdir):
    """
    Test provide_saved_model with no caching
    """
    model_config = {"sklearn.decomposition.pca.PCA": {"svd_solver": "auto"}}
    data_config = get_random_data()
    output_dir = os.path.join(tmpdir, "model")
    machine = Machine(
        name="model-name", dataset=data_config, model=model_config, project_name="test"
    )
    ModelBuilder(machine).build(output_dir=output_dir)

    # Assert the model was saved at the location
    # Should be model file, and the metadata
    assert len(os.listdir(output_dir)) == 2


def test_provide_saved_model_caching_handle_existing_same_dir(tmpdir):
    """If the model exists in the model register, and the path there is the
    same as output_dir, output_dir is returned"""
    model_config = {"sklearn.decomposition.pca.PCA": {"svd_solver": "auto"}}
    data_config = get_random_data()
    output_dir = os.path.join(tmpdir, "model")
    registry_dir = os.path.join(tmpdir, "registry")
    machine = Machine(
        name="model-name", dataset=data_config, model=model_config, project_name="test"
    )
    builder = ModelBuilder(machine)
    builder.build(output_dir=output_dir, model_register_dir=registry_dir)
    assert builder.cached_model_path == output_dir

    # Saving to same output_dir as the one saved in the registry just returns the output_dir
    builder.build(output_dir=output_dir, model_register_dir=registry_dir)
    assert builder.cached_model_path == output_dir


def test_provide_saved_model_caching_handle_existing_different_register(tmpdir):
    """If the model exists in the model register, but the output_dir is not where
    the model is, the model is copied to the new location, unless the new location
    already exists. If it does then return it"""
    model_config = {"sklearn.decomposition.pca.PCA": {"svd_solver": "auto"}}
    data_config = get_random_data()
    output_dir1 = os.path.join(tmpdir, "model1")
    output_dir2 = os.path.join(tmpdir, "model2")

    registry_dir = os.path.join(tmpdir, "registry")
    machine = Machine(
        name="model-name", dataset=data_config, model=model_config, project_name="test"
    )
    builder = ModelBuilder(machine)
    builder.build(output_dir=output_dir1, model_register_dir=registry_dir)

    builder.build(output_dir=output_dir2, model_register_dir=registry_dir)
    assert builder.cached_model_path == output_dir2

    builder.build(output_dir=output_dir2, model_register_dir=registry_dir)
    assert builder.cached_model_path == output_dir2


@pytest.mark.parametrize(
    "should_be_equal,metadata,tag_list,replace_cache",
    [
        (True, None, None, False),
        (True, Metadata(user_defined={"metadata": "something"}), None, False),
        (False, Metadata(user_defined={"metadata": "something"}), None, True),
        (False, None, [SensorTag("extra_tag", None)], False),
        (False, None, None, True),  # replace_cache gives a new model location
    ],
)
def test_provide_saved_model_caching(
    should_be_equal: bool,
    metadata: Optional[Metadata],
    tag_list: Optional[List[SensorTag]],
    replace_cache,
    tmpdir,
):
    """
    Test provide_saved_model with caching and possible cache busting if tag_list, or replace_cache is set.

    Builds two models and checks if their model-creation-date's are the same,
    which will be if and only if there is caching.

    Parameters
    ----------
    should_be_equal : bool
        Do we expect the two generated models to be at the same location or not? I.e. do
        we expect caching.
    metadata: Metadata
        Optional metadata which will be used as metadata for the second model.
    tag_list
        Optional list of strings which be used as the taglist in the dataset for the
        second model.
    replace_cache: bool
        Should we force a model cache replacement?

    """

    if tag_list is None:
        tag_list = []
    if metadata is None:
        metadata = Metadata()

    model_config = {"sklearn.decomposition.pca.PCA": {"svd_solver": "auto"}}
    data_config = get_random_data()
    output_dir = os.path.join(tmpdir, "model")
    registry_dir = os.path.join(tmpdir, "registry")
    machine = Machine(
        name="model-name", dataset=data_config, model=model_config, project_name="test"
    )
    _, first_machine = ModelBuilder(machine).build(
        output_dir=output_dir, model_register_dir=registry_dir
    )

    if tag_list:
        data_config["tag_list"] = tag_list

    new_output_dir = os.path.join(tmpdir, "model2")
    _, second_machine = ModelBuilder(
        machine=Machine(
            name="model-name",
            dataset=data_config,
            model=model_config,
            metadata=metadata,
            project_name="test",
            runtime={"something": True},
        )
    ).build(
        output_dir=new_output_dir,
        model_register_dir=registry_dir,
        replace_cache=replace_cache,
    )

    model1_creation_date = (
        first_machine.metadata.build_metadata.model.model_creation_date
    )
    model2_creation_date = (
        second_machine.metadata.build_metadata.model.model_creation_date
    )
    assert "something" in second_machine.runtime

    if should_be_equal:
        assert model1_creation_date == model2_creation_date
    else:
        assert model1_creation_date != model2_creation_date

    if metadata is not None:
        assert metadata.user_defined == second_machine.metadata.user_defined


@pytest.mark.parametrize(
    "metrics_",
    (
        ["sklearn.metrics.r2_score"],
        ["r2_score"],  # string names for funcs in sklearn.metrics should also work
        None,
        ["sklearn.metrics.r2_score", "sklearn.metrics.explained_variance_score"],
    ),
)
def test_model_builder_metrics_list(metrics_: Optional[List[str]]):
    model_config = {
        "sklearn.multioutput.MultiOutputRegressor": {
            "estimator": "sklearn.linear_model.LinearRegression"
        }
    }
    data_config = get_random_data()

    evaluation_config: Dict[str, Any] = {"cv_mode": "full_build"}
    if metrics_:
        evaluation_config.update({"metrics": metrics_})

    machine = Machine(
        name="model-name",
        dataset=data_config,
        model=model_config,
        evaluation=evaluation_config,
        project_name="test",
    )
    _model, machine = ModelBuilder(machine).build()

    expected_metrics = metrics_ or [
        "sklearn.metrics.explained_variance_score",
        "sklearn.metrics.r2_score",
        "sklearn.metrics.mean_squared_error",
        "sklearn.metrics.mean_absolute_error",
    ]

    assert all(
        metric.split(".")[-1].replace("_", "-")
        in machine.metadata.build_metadata.model.cross_validation.scores
        for metric in expected_metrics
    )


def test_metrics_from_list():
    """
    Check getting functions from a list of metric names
    """
    default = ModelBuilder.metrics_from_list()
    assert default == [
        metrics.explained_variance_score,
        metrics.r2_score,
        metrics.mean_squared_error,
        metrics.mean_absolute_error,
    ]

    specifics = ModelBuilder.metrics_from_list(
        ["sklearn.metrics.adjusted_mutual_info_score", "sklearn.metrics.r2_score"]
    )
    assert specifics == [metrics.adjusted_mutual_info_score, metrics.r2_score]


@pytest.mark.parametrize("seed", (None, 1234))
@pytest.mark.parametrize(
    "model_config",
    (
        {
            "sklearn.multioutput.MultiOutputRegressor": {
                "estimator": "sklearn.ensemble.forest.RandomForestRegressor"
            }
        },
        {
            "gordo.machine.model.models.KerasAutoEncoder": {
                "kind": "feedforward_hourglass"
            }
        },
    ),
)
def test_setting_seed(seed, model_config):
    """
    Test that we can set the seed and get same results.
    """

    data_config = get_random_data()
    evaluation_config = {"cv_mode": "full_build", "seed": seed}

    # Training two instances, without a seed should result in different scores,
    # while doing it with a seed should result in the same scores.
    machine = Machine(
        name="model-name",
        dataset=data_config,
        model=model_config,
        evaluation=evaluation_config,
        project_name="test",
    )
    _model, machine1 = ModelBuilder(machine).build()
    _model, machine2 = ModelBuilder(machine).build()

    df1 = pd.DataFrame.from_dict(
        machine1.metadata.build_metadata.model.cross_validation.scores
    )
    df2 = pd.DataFrame.from_dict(
        machine2.metadata.build_metadata.model.cross_validation.scores
    )

    # Equality depends on the seed being set.
    if seed:
        assert df1.equals(df2)
    else:
        assert not df1.equals(df2)


@pytest.mark.parametrize(
    "cv",
    (
        {"sklearn.model_selection.TimeSeriesSplit": {"n_splits": 5}},
        {
            "sklearn.model_selection.TimeSeriesSplit": {
                "n_splits": 5,
                "max_train_size": 10,
            }
        },
        {"sklearn.model_selection.ShuffleSplit": {"n_splits": 5}},
        None,
    ),
)
@patch("gordo.serializer.from_definition")
def test_n_splits_from_config(mocked_pipeline_from_definition, cv):
    """
    Test that we can set arbitrary splitters and parameters in the config file which is called by the serializer.
    """
    data_config = get_random_data()
    evaluation_config = {"cv_mode": "full_build"}
    if cv:
        evaluation_config["cv"] = cv

    model_config = {
        "sklearn.multioutput.MultiOutputRegressor": {
            "estimator": "sklearn.ensemble.forest.RandomForestRegressor"
        }
    }

    machine = Machine(
        name="model-name",
        dataset=data_config,
        model=model_config,
        evaluation=evaluation_config,
        project_name="test",
    )

    ModelBuilder(machine).build()

    if cv:
        mocked_pipeline_from_definition.assert_called_with(cv)
    else:
        mocked_pipeline_from_definition.assert_called_with(
            {"sklearn.model_selection.TimeSeriesSplit": {"n_splits": 3}}
        )


@patch("gordo.machine.Machine.report")
def test_builder_calls_machine_report(mocked_report_method, metadata):
    """
    When building a machine, the Modelbuilder.build should call Machine.report()
    so that it can run any reporters in the Machine's runtime.
    """
    machine = Machine(**metadata)
    ModelBuilder(machine).build()
    assert mocked_report_method.called_once()



# -*- coding: utf-8 -*-

from gordo.machine.dataset import datasets


def _get_dataset(config):
    """
    Return a GordoBaseDataSet object of a certain type, given a config dict
    """
    dataset_config = dict(config)
    kind = dataset_config.pop("type", "")
    Dataset = getattr(datasets, kind, None)
    if Dataset is None:
        raise ValueError(f'Dataset type "{kind}" is not supported!')

    return Dataset(**dataset_config)



import re
import unittest
from datetime import datetime
from typing import Iterable, List, Pattern, Any

import pandas as pd
import pytest

from gordo.machine.dataset.data_provider.base import GordoBaseDataProvider
from gordo.machine.dataset.data_provider import providers
from gordo.machine.dataset.data_provider.providers import (
    load_series_from_multiple_providers,
)
from gordo.machine.dataset.sensor_tag import SensorTag


class MockProducerRegExp(GordoBaseDataProvider):
    def can_handle_tag(self, tag: SensorTag):
        return self.regexp.match(tag.name)

    def load_series(
        self,
        train_start_date: datetime,
        train_end_date: datetime,
        tag_list: List[SensorTag],
        dry_run=False,
    ) -> Iterable[pd.Series]:
        for tag in tag_list:
            if self.regexp.match(tag.name):
                yield pd.Series(name=str(self.regexp.pattern))
            else:
                raise ValueError(f"Unable to find base path from tag {tag.name}")

    def __init__(self, regexp: Pattern[Any], **kwargs):
        """
        Mock producer which can handle tags which follow the regexp, and yields empty
        dataframes with one column being the regexp pattern.

        Parameters
        ----------
        regexp
            Regular expression for tags the mock producer can accept
        """
        self.regexp = regexp  # type: Pattern[Any]


class LoadMultipleDataFramesTest(unittest.TestCase):
    @classmethod
    def setUp(self):
        # Producer only accepting tags which starts with "ab"
        self.ab_producer = MockProducerRegExp(re.compile("ab.*"))
        # Producer only accepting tags which contain a "b"
        self.containing_b_producer = MockProducerRegExp(re.compile(".*b.*"))

    def test_load_multiple_raises_with_no_matches(self):
        """If no provider matches a tag then load_series_from_multiple_providers
        raises a ValueError when the generator is realized"""
        with self.assertRaises(ValueError):
            list(
                load_series_from_multiple_providers(
                    [self.ab_producer, self.containing_b_producer],
                    None,
                    None,
                    [
                        SensorTag("ab", None),
                        SensorTag("tag_not_matching_any_of_the_regexps", None),
                    ],
                )
            )

    def test_load_multiple_matches_loads_from_first(self):
        """When a tag can be read from multiple providers it is the first provider in
        the list of providers which gets the job"""
        series_collection = list(
            load_series_from_multiple_providers(
                [self.ab_producer, self.containing_b_producer],
                None,
                None,
                [SensorTag("abba", None)],
            )
        )
        self.assertEqual(series_collection[0].name, "ab.*")

    def test_load_from_multiple_providers(self):
        """ Two tags, each belonging to different data producers, and both gets loaded
        """
        series_collection = list(
            load_series_from_multiple_providers(
                [self.ab_producer, self.containing_b_producer],
                None,
                None,
                [SensorTag("abba", None), SensorTag("cba", None)],
            )
        )
        self.assertEqual(series_collection[0].name, "ab.*")
        self.assertEqual(series_collection[1].name, ".*b.*")


@pytest.mark.parametrize(
    "provider,expected_params",
    (
        (
            providers.RandomDataProvider(200, max_size=205),
            {"min_size": 200, "max_size": 205},
        ),
        (
            providers.InfluxDataProvider("measurement", value_name="Value"),
            {"measurement": "measurement", "value_name": "Value"},
        ),
    ),
)
def test_data_provider_serializations(
    provider: GordoBaseDataProvider, expected_params: dict
):
    """
    Test a given provider can be serialized to dict and back
    """

    encoded = provider.to_dict()

    # Verify the expected parameter kwargs match
    for k, v in expected_params.items():
        assert encoded[k] == v

    # Should have inserted the name of the class as 'type'
    assert provider.__class__.__name__ == encoded["type"]

    # Should be able to recreate the object from encoded directly
    cloned = provider.__class__.from_dict(encoded)
    assert type(cloned) == type(provider)



# -*- coding: utf-8 -*-
import collections
import re
import os
import logging
from typing import AnyStr

from azure.datalake.store import core, lib
from typing.re import Pattern

logger = logging.getLogger(__name__)


def get_datalake_token(
    interactive: bool = False, dl_service_auth_str: str = None
) -> lib.DataLakeCredential:
    """
    Provides a token for azure datalake, either by parsing a datalake service
    authentication string or using interactive authentication

    Parameters
    ----------
    interactive: bool
        If true then use interactive authentication
    dl_service_auth_str: str
        String on the format tenant:client_id:client_secret

    Returns
    -------
    lib.DataLakeCredential
        A lib.DataLakeCredential which can be used to authenticate towards the datalake
    """
    dl_service_auth_str = (
        os.environ.get("DL_SERVICE_AUTH_STR")
        if str(dl_service_auth_str) == "None"
        else dl_service_auth_str
    )
    if interactive:
        logger.info("Attempting to use interactive azure authentication")
        return lib.auth()
    elif dl_service_auth_str:
        logger.info(f"Attempting to use datalake service authentication")
        dl_service_auth_elems = dl_service_auth_str.split(":")
        tenant = dl_service_auth_elems[0]
        client_id = dl_service_auth_elems[1]
        client_secret = dl_service_auth_elems[2]
        token = lib.auth(
            tenant_id=tenant,
            client_secret=client_secret,
            client_id=client_id,
            resource="https://datalake.azure.net/",
        )
        return token
    else:
        raise ValueError(
            f"Either interactive (value: {interactive}) must be True, "
            f"or dl_service_auth_str (value: {dl_service_auth_str}) "
            "must be set. "
        )


def create_adls_client(
    storename: str, dl_service_auth_str: str = None, interactive: bool = False
) -> core.AzureDLFileSystem:
    """
    Creates an ADLS file system client.

    Parameters
    ----------
    storename: str
        Name of datalake store.
    dl_service_auth_str: str
        Authentication string to use
    interactive: bool
        If true then use interactive authentication


    Returns
    -------
    core.AzureDLFileSystem
        Instance of AzureDLFileSystem, ready to use
    """
    token = get_datalake_token(
        interactive=interactive, dl_service_auth_str=dl_service_auth_str
    )

    adls_file_system_client = core.AzureDLFileSystem(token, store_name=storename)
    return adls_file_system_client


def walk_azure(
    client: core.AzureDLFileSystem,
    base_path: str,
    include_regexp: Pattern[AnyStr] = re.compile(".*"),
    exclude_regexp: Pattern[AnyStr] = re.compile("a^"),
):
    """
    Walks azure datalake `azure_data_store` in a depth-first fashion from
    `base_path`, yielding files which match `include_re` AND not match `exclude_re`

    Notes
    -----
    If `base_path` does not exist then the generator returns nothing, it does not fail.

    """
    logger.info(
        "Starting a azure walk with base_path: {} with include_regexp: {} "
        "and exclude_regexp: {}".format(base_path, include_regexp, exclude_regexp)
    )

    adls_file_system_client = client
    if not adls_file_system_client.exists(base_path):
        return
    fringe = collections.deque(adls_file_system_client.ls(base_path, detail=True))
    while fringe:
        a_path = fringe.pop()
        if a_path["type"] == "DIRECTORY":
            logger.info("Expanding the directory %s" % a_path["name"])
            fringe.extend(adls_file_system_client.ls(a_path["name"], detail=True))
        if a_path["type"] == "FILE":
            file_path = a_path["name"]
            if include_regexp.match(file_path) and not exclude_regexp.match(file_path):
                logger.info("Returning the file_path %s" % file_path)
                yield file_path
            else:
                logger.info(
                    f"Found that the file_path {file_path} does not satisfy regexp "
                    f"requirements"
                )


def is_file(client: core.AzureDLFileSystem, path: str):
    try:
        info = client.info(path)
    except FileNotFoundError:
        return False
    return info["type"] == "FILE"



# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd

from typing import Optional, Union
from datetime import timedelta

from sklearn.preprocessing import RobustScaler
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.model_selection import TimeSeriesSplit, cross_validate

from gordo.machine.model.base import GordoBase
from gordo.machine.model import utils as model_utils
from gordo.machine.model.models import KerasAutoEncoder
from gordo.machine.model.anomaly.base import AnomalyDetectorBase


class DiffBasedAnomalyDetector(AnomalyDetectorBase):
    def __init__(
        self,
        base_estimator: BaseEstimator = KerasAutoEncoder(kind="feedforward_hourglass"),
        scaler: TransformerMixin = RobustScaler(),
        require_thresholds: bool = True,
    ):
        """
        Classifier which wraps a ``base_estimator`` and provides a diff error
        based approach to anomaly detection.

        It trains a ``scaler`` to the target **after** training, purely for
        error calculations. The underlying ``base_estimator`` is trained
        with the original, unscaled, ``y``.

        Parameters
        ----------
        base_estimator: sklearn.base.BaseEstimator
            The model to which normal ``.fit``, ``.predict`` methods will be used.
            defaults to py:class:`gordo.machine.model.models.KerasAutoEncoder` with
            ``kind='feedforward_hourglass``
        scaler: sklearn.base.TransformerMixn
            Defaults to ``sklearn.preprocessing.RobustScaler``
            Used for transforming model output and the original ``y`` to calculate
            the difference/error in model output vs expected.
        require_thresholds: bool
            Requires calculating ``thresholds_`` via a call to :func:`~DiffBasedAnomalyDetector.cross_validate`.
            If this is set (default True), but :func:`~DiffBasedAnomalyDetector.cross_validate`
            was not called before calling :func:`~DiffBasedAnomalyDetector.anomaly` an ``AttributeError``
            will be raised.
        """
        self.base_estimator = base_estimator
        self.scaler = scaler
        self.require_thresholds = require_thresholds

    def __getattr__(self, item):
        """
        Treat this model as transparent into base_estimator unless
        referring to something owned by this object
        """
        if item in self.__dict__:
            return getattr(self, item)
        else:
            return getattr(self.base_estimator, item)

    def get_metadata(self):
        metadata = dict()

        if hasattr(self, "feature_thresholds_"):
            metadata["feature-thresholds"] = self.feature_thresholds_.tolist()
        if hasattr(self, "aggregate_threshold_"):
            metadata["aggregate-threshold"] = self.aggregate_threshold_
        if hasattr(self, "feature_thresholds_per_fold_"):
            metadata[
                "feature-thresholds-per-fold"
            ] = self.feature_thresholds_per_fold_.to_dict()
        if hasattr(self, "aggregate_thresholds_per_fold_"):
            metadata[
                "aggregate-thresholds-per-fold"
            ] = self.aggregate_thresholds_per_fold_
        if isinstance(self.base_estimator, GordoBase):
            metadata.update(self.base_estimator.get_metadata())
        else:
            metadata.update(
                {"scaler": str(self.scaler), "base_estimator": str(self.base_estimator)}
            )
        return metadata

    def score(
        self,
        X: Union[np.ndarray, pd.DataFrame],
        y: Union[np.ndarray, pd.DataFrame],
        sample_weight: Optional[np.ndarray] = None,
    ) -> float:
        return self.base_estimator.score(X, y)

    def get_params(self, deep=True):
        return {"base_estimator": self.base_estimator, "scaler": self.scaler}

    def fit(self, X: np.ndarray, y: np.ndarray):
        self.base_estimator.fit(X, y)
        self.scaler.fit(y)  # Scaler is used for calculating errors in .anomaly()
        return self

    def cross_validate(
        self,
        *,
        X: Union[pd.DataFrame, np.ndarray],
        y: Union[pd.DataFrame, np.ndarray],
        cv=TimeSeriesSplit(n_splits=3),
        **kwargs,
    ):
        """
        Run cross validation on the model, and will update the model's threshold value based
        on the cross validation folds

        Parameters
        ----------
        X: Union[pd.DataFrame, np.ndarray]
            Input data to the model
        y: Union[pd.DataFrame, np.ndarray]
            Target data
        kwargs: dict
            Any additional kwargs to be passed to :func:`sklearn.model_selection.cross_validate`

        Returns
        -------
        dict
        """
        # Depend on having the trained fold models
        kwargs.update(dict(return_estimator=True, cv=cv))

        cv_output = cross_validate(self, X=X, y=y, **kwargs)

        self.feature_thresholds_per_fold_ = pd.DataFrame()
        self.aggregate_thresholds_per_fold_ = {}

        for i, ((train_idxs, test_idxs), split_model) in enumerate(
            zip(kwargs["cv"].split(X, y), cv_output["estimator"])
        ):
            y_pred = split_model.predict(
                X.iloc[test_idxs] if isinstance(X, pd.DataFrame) else X[test_idxs]
            )

            # Adjust y_true for any possible model offset in its prediction
            test_idxs = test_idxs[-len(y_pred) :]
            y_true = y.iloc[test_idxs] if isinstance(y, pd.DataFrame) else y[test_idxs]

            # Model's timestep scaled mse over all features
            scaled_mse = self._scaled_mse_per_timestep(split_model, y_true, y_pred)

            # For the aggregate threshold for the fold model, use the mse of scaled residuals per timestep
            aggregate_threshold_fold = scaled_mse.rolling(6).min().max()
            self.aggregate_thresholds_per_fold_[f"fold-{i}"] = aggregate_threshold_fold

            # Accumulate the rolling mins of diffs into common df
            tag_thresholds_fold = (
                pd.DataFrame(np.abs(y_pred - y_true)).rolling(6).min().max()
            )
            tag_thresholds_fold.name = f"fold-{i}"
            self.feature_thresholds_per_fold_ = self.feature_thresholds_per_fold_.append(
                tag_thresholds_fold
            )

        # Final thresholds are the thresholds from the last cv split/fold
        self.feature_thresholds_ = tag_thresholds_fold

        # For the aggregate also use the thresholds from the last split/fold
        self.aggregate_threshold_ = aggregate_threshold_fold

        return cv_output

    @staticmethod
    def _scaled_mse_per_timestep(
        model: BaseEstimator,
        y_true: Union[pd.DataFrame, np.ndarray],
        y_pred: Union[pd.DataFrame, np.ndarray],
    ) -> pd.Series:
        """
        Calculate the scaled MSE per timestep/sample

        Parameters
        ----------
        model: BaseEstimator
            Instance of a fitted :class:`~DiffBasedAnomalyDetector`
        y_true: Union[numpy.ndarray, pd.DataFrame]
        y_pred: Union[numpy.ndarray, pd.DataFrame]

        Returns
        -------
        panadas.Series
            The MSE calculated from the scaled y and y predicted.
        """
        scaled_y_true = model.scaler.transform(y_true)
        scaled_y_pred = model.scaler.transform(y_pred)
        mse_per_time_step = ((scaled_y_pred - scaled_y_true) ** 2).mean(axis=1)
        return pd.Series(mse_per_time_step)

    def anomaly(
        self, X: pd.DataFrame, y: pd.DataFrame, frequency: Optional[timedelta] = None
    ) -> pd.DataFrame:
        """
        Create an anomaly dataframe from the base provided dataframe.

        Parameters
        ----------
        X: pd.DataFrame
            Dataframe representing the data to go into the model.
        y: pd.DataFrame
            Dataframe representing the target output of the model.

        Returns
        -------
        pd.DataFrame
            A superset of the original base dataframe with added anomaly specific
            features
        """

        # Get the model output, falling back to transform if 'predict' doesn't exist
        model_output = (
            self.predict(X) if hasattr(self, "predict") else self.transform(X)
        )

        # Create the basic dataframe with 'model-output' & 'model-input'
        data = model_utils.make_base_dataframe(
            tags=X.columns,
            model_input=getattr(X, "values", X),
            model_output=model_output,
            target_tag_list=y.columns,
            index=getattr(X, "index", None),
            frequency=frequency,
        )

        model_out_scaled = pd.DataFrame(
            self.scaler.transform(data["model-output"]),
            columns=data["model-output"].columns,
            index=data.index,
        )

        # Calculate the absolute scaled tag anomaly
        # Ensure to offset the y to match model out, which could be less if it was an LSTM
        scaled_y = self.scaler.transform(y)
        tag_anomaly_scaled = np.abs(model_out_scaled - scaled_y[-len(data) :, :])
        tag_anomaly_scaled.columns = pd.MultiIndex.from_product(
            (("tag-anomaly-scaled",), tag_anomaly_scaled.columns)
        )
        data = data.join(tag_anomaly_scaled)

        # Calculate scaled total anomaly
        data["total-anomaly-scaled"] = np.square(data["tag-anomaly-scaled"]).mean(
            axis=1
        )

        # Calculate the absolute unscaled tag anomalies
        unscaled_abs_diff = pd.DataFrame(
            data=np.abs(data["model-output"].values - y.values[-len(data) :, :]),
            index=data.index,
            columns=pd.MultiIndex.from_product(
                (("tag-anomaly-unscaled",), y.columns.tolist())
            ),
        )
        data = data.join(unscaled_abs_diff)

        # Calculate the scaled total anomaly
        data["total-anomaly-unscaled"] = np.square(data["tag-anomaly-unscaled"]).mean(
            axis=1
        )

        # If we have `thresholds_` values, then we can calculate anomaly confidence
        if hasattr(self, "feature_thresholds_"):
            confidence = tag_anomaly_scaled.values / self.feature_thresholds_.values

            # Dataframe of % abs_diff is of the thresholds
            anomaly_confidence_scores = pd.DataFrame(
                confidence,
                index=tag_anomaly_scaled.index,
                columns=pd.MultiIndex.from_product(
                    (("anomaly-confidence",), data["model-output"].columns)
                ),
            )
            data = data.join(anomaly_confidence_scores)

        if hasattr(self, "aggregate_threshold_"):
            data["total-anomaly-confidence"] = (
                data["total-anomaly-scaled"] / self.aggregate_threshold_
            )

        # Explicitly raise error if we were required to do threshold based calculations
        # should would have required a call to .cross_validate before .anomaly
        if self.require_thresholds and not any(
            hasattr(self, attr)
            for attr in ("feature_thresholds_", "aggregate_threshold_")
        ):
            raise AttributeError(
                f"`require_thresholds={self.require_thresholds}` however `.cross_validate`"
                f"needs to be called in order to calculate these thresholds before calling `.anomaly`"
            )

        return data



# -*- coding: utf-8 -*-

import asyncio
import io
import json
import logging
import os
import re
import tempfile
from threading import Lock
from typing import List
from unittest.mock import patch

import docker
from flask import Request
import pytest
import responses

from gordo import serializer
from gordo.machine.dataset.sensor_tag import SensorTag

from gordo.server import server
from gordo.builder.local_build import local_build
from gordo.machine.dataset import sensor_tag
from gordo.machine.dataset.sensor_tag import to_list_of_strings
from gordo.server import server as gordo_ml_server

from tests import utils as tu

logger = logging.getLogger(__name__)

TEST_SERVER_MUTEXT = Lock()


def pytest_collection_modifyitems(items):
    """
    Update all tests which use influxdb to be marked as a dockertest
    """
    for item in items:
        if hasattr(item, "fixturenames") and "influxdb" in item.fixturenames:
            item.add_marker(pytest.mark.dockertest)


@pytest.fixture(autouse=True)
def check_event_loop():
    loop = asyncio.get_event_loop()
    if loop.is_closed():
        logger.info("Creating new event loop!")
        asyncio.set_event_loop(asyncio.new_event_loop())


@pytest.fixture(scope="session")
def gordo_host():
    return "localhost"


@pytest.fixture(scope="session")
def gordo_project():
    return "gordo-test"


@pytest.fixture(scope="session")
def gordo_name():
    return "machine-1"


@pytest.fixture(scope="session")
def gordo_single_target(gordo_name):
    return gordo_name


@pytest.fixture(scope="session")
def gordo_targets(gordo_single_target):
    return [gordo_single_target]


@pytest.fixture(scope="session")
def sensors():
    return [SensorTag(f"tag-{i}", None) for i in range(4)]


@pytest.fixture(scope="session")
def sensors_str(sensors):
    return to_list_of_strings(sensors)


@pytest.fixture(scope="session")
def influxdb_name():
    return "testdb"


@pytest.fixture(scope="session")
def influxdb_user():
    return "root"


@pytest.fixture(scope="session")
def influxdb_password():
    return "root"


@pytest.fixture(scope="session")
def influxdb_measurement():
    return "sensors"


@pytest.fixture(scope="session")
def influxdb_fixture_args(sensors_str, influxdb_name, influxdb_user, influxdb_password):
    return (sensors_str, influxdb_name, influxdb_user, influxdb_password, sensors_str)


@pytest.fixture(scope="session")
def influxdb_uri(influxdb_user, influxdb_password, influxdb_name):
    return f"{influxdb_user}:{influxdb_password}@localhost:8086/{influxdb_name}"


@pytest.fixture(scope="session")
def gordo_revision():
    return "1234"


@pytest.fixture(scope="session")
def api_version():
    return "v0"


@pytest.fixture(scope="session")
def base_route(api_version, gordo_project, gordo_name):
    return f"/gordo/{api_version}/{gordo_project}/{gordo_name}"


@pytest.fixture(scope="session")
def model_collection_directory(gordo_revision: str):
    with tempfile.TemporaryDirectory() as tmp_dir:
        collection_dir = os.path.join(tmp_dir, gordo_revision)
        os.makedirs(collection_dir, exist_ok=True)
        yield collection_dir


@pytest.fixture(scope="session")
def config_str(gordo_name: str, sensors: List[SensorTag]):
    """
    Fixture: Default config for testing
    """
    return f"""
            machines:
              - dataset:
                  tags:
                    - {sensors[0].name}
                    - {sensors[1].name}
                    - {sensors[2].name}
                    - {sensors[3].name}
                  target_tag_list:
                    - {sensors[0].name}
                    - {sensors[1].name}
                    - {sensors[2].name}
                    - {sensors[3].name}
                  train_start_date: '2019-01-01T00:00:00+00:00'
                  train_end_date: '2019-10-01T00:00:00+00:00'
                  asset: asgb
                  data_provider:
                    type: RandomDataProvider
                metadata:
                  information: Some sweet information about the model
                model:
                  gordo.machine.model.anomaly.diff.DiffBasedAnomalyDetector:
                    require_thresholds: false
                    base_estimator:
                      sklearn.pipeline.Pipeline:
                        steps:
                        - sklearn.preprocessing.data.MinMaxScaler
                        - gordo.machine.model.models.KerasAutoEncoder:
                            kind: feedforward_hourglass
                name: {gordo_name}
             """


@pytest.fixture(scope="session")
def trained_model_directory(
    model_collection_directory: str, config_str: str, gordo_name: str
):
    """
    Fixture: Train a basic AutoEncoder and save it to a given directory
    will also save some metadata with the model
    """

    # Model specific to the model being trained here
    model_dir = os.path.join(model_collection_directory, gordo_name)
    os.makedirs(model_dir, exist_ok=True)

    builder = local_build(config_str=config_str)
    model, metadata = next(builder)  # type: ignore
    serializer.dump(model, model_dir, metadata=metadata.to_dict())
    yield model_dir


@pytest.fixture
def metadata(trained_model_directory):
    return serializer.load_metadata(trained_model_directory)


@pytest.fixture(scope="session")
def gordo_ml_server_client(
    request, model_collection_directory, trained_model_directory
):

    with tu.temp_env_vars(MODEL_COLLECTION_DIR=model_collection_directory):

        app = server.build_app()
        app.testing = True

        # always return a valid asset for any tag name
        with patch.object(sensor_tag, "_asset_from_tag_name", return_value="default"):
            yield app.test_client()


@pytest.yield_fixture
def postgresdb():
    client = docker.from_env()
    postgres = client.containers.run(
        image="postgres:11-alpine",
        environment={"POSTGRES_USER": "postgres", "POSTGRES_PASSWORD": "postgres"},
        ports={"5432/tcp": "5432"},
        remove=True,
        detach=True,
    )
    import time

    time.sleep(5)
    yield
    postgres.kill()


@pytest.fixture(scope="session")
def base_influxdb(
    sensors, influxdb_name, influxdb_user, influxdb_password, influxdb_measurement
):
    """
    Fixture to yield a running influx container and pass a tests.utils.InfluxDB
    object which can be used to reset the db to it's original data state.
    """
    client = docker.from_env()

    logger.info("Starting up influx!")
    influx = None
    try:
        influx = client.containers.run(
            image="influxdb:1.7-alpine",
            environment={
                "INFLUXDB_DB": influxdb_name,
                "INFLUXDB_ADMIN_USER": influxdb_user,
                "INFLUXDB_ADMIN_PASSWORD": influxdb_password,
            },
            ports={"8086/tcp": "8086"},
            remove=True,
            detach=True,
        )
        if not tu.wait_for_influx(influx_host="localhost:8086"):
            raise TimeoutError("Influx failed to start")

        logger.info(f"Started influx DB: {influx.name}")

        # Create the interface to the running instance, set default state, and yield it.
        db = tu.InfluxDB(
            sensors,
            influxdb_name,
            influxdb_user,
            influxdb_password,
            influxdb_measurement,
        )
        db.reset()
        logger.info("STARTED INFLUX INSTANCE")
        yield db

    finally:
        logger.info("Killing influx container")
        if influx:
            influx.kill()
        logger.info("Killed influx container")


@pytest.fixture
def influxdb(base_influxdb):
    """
    Fixture to take a running influx and do a reset after each test to ensure
    the data state is the same for each test.
    """
    logger.info("DOING A RESET ON INFLUX DATA")
    base_influxdb.reset()


@pytest.fixture(scope="session")
def argo_version(repo_dir):
    with open(os.path.join(repo_dir, "Dockerfile-GordoDeploy")) as f:
        match = next(re.finditer(r'ARGO_VERSION="(\w\d+.\d+.\d+)"', f.read()), None)
    if match is None:
        raise LookupError(
            "Failed to determine argo version from Dockerfile-GordoDeploy"
        )
    return match.groups()[0]


@pytest.fixture(scope="module")
def ml_server(
    model_collection_directory, trained_model_directory, gordo_host, gordo_project
):
    """
    # TODO: This is bananas, make into a proper object with context support?

    Mock a deployed controller deployment

    Parameters
    ----------
    gordo_host: str
        Host controller should pretend to run on
    gordo_project: str
        Project controller should pretend to care about
    model_collection_directory: str
        Directory of the model to use in the target(s)

    Returns
    -------
    None
    """
    with tu.temp_env_vars(MODEL_COLLECTION_DIR=model_collection_directory):
        # Create gordo ml servers
        gordo_server_app = gordo_ml_server.build_app()
        gordo_server_app.testing = True
        gordo_server_app = gordo_server_app.test_client()

        def gordo_ml_server_callback(request):
            """
            Redirect calls to a gordo server to reflect what the local testing app gives
            will call the correct path (assuminng only single level paths) on the
            gordo app.
            """
            if request.method in ("GET", "POST"):

                kwargs = dict()
                if request.body:
                    flask_request = Request.from_values(
                        content_length=len(request.body),
                        input_stream=io.BytesIO(request.body),
                        content_type=request.headers["Content-Type"],
                        method=request.method,
                    )
                    if flask_request.json:
                        kwargs["json"] = flask_request.json
                    else:
                        kwargs["data"] = {
                            k: (io.BytesIO(f.read()), f.filename)
                            for k, f in flask_request.files.items()
                        }

                with TEST_SERVER_MUTEXT:
                    resp = getattr(gordo_server_app, request.method.lower())(
                        request.path_url, headers=dict(request.headers), **kwargs
                    )
                return (
                    resp.status_code,
                    resp.headers,
                    json.dumps(resp.json) if resp.json is not None else resp.data,
                )

        with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
            rsps.add_callback(
                responses.GET,
                re.compile(rf".*{gordo_host}.*\/gordo\/v0\/{gordo_project}\/.+"),
                callback=gordo_ml_server_callback,
                content_type="application/json",
            )
            rsps.add_callback(
                responses.POST,
                re.compile(rf".*{gordo_host}.*\/gordo\/v0\/{gordo_project}\/.*.\/.*"),
                callback=gordo_ml_server_callback,
                content_type="application/json",
            )

            rsps.add_passthru("http+docker://")  # Docker
            rsps.add_passthru("http://localhost:8086")  # Local influx
            rsps.add_passthru("http://localhost:8087")  # Local influx

            yield


@pytest.fixture(scope="session")
def repo_dir():
    """
    Return the repository directory for gordo infrastructure
    """
    return os.path.join(os.path.dirname(__file__), "..")



import pytest

from gordo.machine.dataset.sensor_tag import (
    normalize_sensor_tags,
    SensorTagNormalizationError,
)
from gordo.machine.dataset.sensor_tag import SensorTag


NON_RESOLVABLE_TAG_NAME1 = "MyBeautifulTag1"
NON_RESOLVABLE_TAG_NAME2 = "MyBeautifulTag2"
asset_nonsense = "ImaginaryAsset"


@pytest.mark.parametrize(
    "good_input_tags,asset,default_asset,expected_output_tags",
    [
        (
            [
                {"name": NON_RESOLVABLE_TAG_NAME1, "asset": asset_nonsense},
                {"name": NON_RESOLVABLE_TAG_NAME2, "asset": asset_nonsense},
            ],
            "ThisAssetCodeWillBeIgnoredSinceAssetMustBeProvidedInDict",
            "AsWillThisSinceAssetMustBeProvidedInDict",
            [
                SensorTag(NON_RESOLVABLE_TAG_NAME1, asset_nonsense),
                SensorTag(NON_RESOLVABLE_TAG_NAME2, asset_nonsense),
            ],
        ),
        (
            ["TRC-123", "GRA-214", "ASGB-212"],
            "ThisWillBeTheAsset",
            "AndThisWillBeIgnoredSinceAllTagsAreResolvable",
            [
                SensorTag("TRC-123", "ThisWillBeTheAsset"),
                SensorTag("GRA-214", "ThisWillBeTheAsset"),
                SensorTag("ASGB-212", "ThisWillBeTheAsset"),
            ],
        ),
        (
            ["TRC-123", "GRA-214", "ASGB-212"],
            None,  # Will deduce asset
            None,
            [
                SensorTag("TRC-123", "1776-troc"),
                SensorTag("GRA-214", "1755-gra"),
                SensorTag("ASGB-212", "1191-asgb"),
            ],
        ),
        (
            [
                SensorTag("TRC-123", "1776-troc"),
                SensorTag("GRA-214", "1755-gra"),
                SensorTag("ASGB-212", "1191-asgb"),
            ],
            "CouldWriteAssetHereButWeDontCare",
            "WillBeIgnoredSinceInputIsSensorTags",
            [
                SensorTag("TRC-123", "1776-troc"),
                SensorTag("GRA-214", "1755-gra"),
                SensorTag("ASGB-212", "1191-asgb"),
            ],
        ),
        (
            [
                [NON_RESOLVABLE_TAG_NAME1, asset_nonsense],
                [NON_RESOLVABLE_TAG_NAME2, asset_nonsense],
            ],
            "WillBeIgnoredSinceInputIsList",
            "WillBeIgnoredSinceInputIsList",
            [
                SensorTag(NON_RESOLVABLE_TAG_NAME1, asset_nonsense),
                SensorTag(NON_RESOLVABLE_TAG_NAME2, asset_nonsense),
            ],
        ),
        (
            ["TRC-123", NON_RESOLVABLE_TAG_NAME2, "ASGB-212"],
            None,
            "fallbackkasset",
            [
                SensorTag("TRC-123", "1776-troc"),
                SensorTag(NON_RESOLVABLE_TAG_NAME2, "fallbackkasset"),
                SensorTag("ASGB-212", "1191-asgb"),
            ],
        ),
    ],
)
def test_normalize_sensor_tags_ok(
    good_input_tags, asset, default_asset, expected_output_tags
):
    tag_list_as_list_of_sensor_tag = normalize_sensor_tags(
        good_input_tags, asset, default_asset=default_asset
    )
    assert tag_list_as_list_of_sensor_tag == expected_output_tags


def test_normalize_sensor_tags_not_ok():
    with pytest.raises(SensorTagNormalizationError):
        tag_list_as_list_of_strings_nonsense = [
            NON_RESOLVABLE_TAG_NAME1,
            NON_RESOLVABLE_TAG_NAME2,
        ]
        normalize_sensor_tags(tag_list_as_list_of_strings_nonsense)



# -*- coding: utf-8 -*-

import logging
import unittest
import yaml
import copy
import pydoc

import pytest
import numpy as np

from sklearn.decomposition import PCA, TruncatedSVD
from sklearn.pipeline import FeatureUnion, Pipeline
from sklearn.preprocessing import MinMaxScaler, FunctionTransformer
from sklearn.multioutput import MultiOutputRegressor

from gordo import serializer
from gordo.serializer import from_definition
import gordo.machine.model.transformer_funcs.general
from gordo.machine.model.register import register_model_builder

logger = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "definition",
    [
        """ 
    sklearn.multioutput.MultiOutputRegressor:
      estimator: sklearn.ensemble.forest.RandomForestRegressor
    """,
        """ 
    sklearn.multioutput.MultiOutputRegressor:
      estimator: 
        sklearn.ensemble.forest.RandomForestRegressor:
          n_estimators: 20
    """,
        """ 
    sklearn.multioutput.MultiOutputRegressor:
      estimator: 
        sklearn.pipeline.Pipeline:
            steps:
                - sklearn.ensemble.forest.RandomForestRegressor:
                    n_estimators: 20
    """,
        """
        sklearn.multioutput.MultiOutputRegressor:
            estimator: 
                sklearn.pipeline.Pipeline:
                    steps:
                        - sklearn.cluster.FeatureAgglomeration:
                            n_clusters: 2
                            pooling_func: numpy.mean
                        - sklearn.linear_model.LinearRegression
        """,
    ],
)
def test_load_from_definition(definition):
    """
    Ensure serializer can load models which take other models as parameters.
    """
    X, y = np.random.random((10, 10)), np.random.random((10, 2))
    definition = yaml.load(definition, Loader=yaml.SafeLoader)
    model = serializer.from_definition(definition)
    assert isinstance(model, MultiOutputRegressor)
    model.fit(X, y)
    model.predict(X)


class ConfigToScikitLearnPipeTestCase(unittest.TestCase):
    def setup_gen(self):
        self.factories = register_model_builder.factories
        for model in self.factories.keys():
            for model_kind in self.factories[model].keys():
                templates = [
                    # This has full parameter names define
                    (
                        f"""
                    sklearn.pipeline.Pipeline:
                        steps:
                            - sklearn.decomposition.pca.PCA:
                                n_components: 2
                                copy: true
                                whiten: false
                                svd_solver:  auto
                                tol: 0.0
                                iterated_power: auto
                                random_state:
                            - sklearn.preprocessing._function_transformer.FunctionTransformer:
                                func: gordo.machine.model.transformer_funcs.general.multiply_by
                                kw_args:
                                    factor: 1
                            - sklearn.pipeline.FeatureUnion:
                                transformer_list:
                                - sklearn.decomposition.pca.PCA:
                                    n_components: 3
                                    copy: true
                                    whiten: false
                                    svd_solver: auto
                                    tol: 0.0
                                    iterated_power: auto
                                    random_state:
                                - sklearn.pipeline.Pipeline:
                                    steps:
                                    - sklearn.preprocessing.data.MinMaxScaler:
                                        feature_range:
                                        - 0
                                        - 1
                                        copy: true
                                    - sklearn.decomposition.truncated_svd.TruncatedSVD:
                                        n_components: 2
                                        algorithm: randomized
                                        n_iter: 5
                                        random_state:
                                        tol: 0.0
                                    memory:
                                n_jobs: 1
                                transformer_weights:
                            - gordo.machine.model.models.{model}: 
                                kind: {model_kind}
                    """,
                        pydoc.locate(f"gordo.machine.model.models.{model}"),
                        model_kind,
                    ),
                    (
                        f"""
                    sklearn.pipeline.Pipeline:
                        steps:
                            - sklearn.decomposition.pca.PCA:
                                n_components: 2
                            - sklearn.preprocessing._function_transformer.FunctionTransformer:
                                func: gordo.machine.model.transformer_funcs.general.multiply_by
                                kw_args:
                                    factor: 1
                            - sklearn.pipeline.FeatureUnion:
                                - sklearn.decomposition.pca.PCA:
                                    n_components: 3
                                - sklearn.pipeline.Pipeline:
                                    - sklearn.preprocessing.data.MinMaxScaler:
                                        feature_range: [0, 1]
                                    - sklearn.decomposition.truncated_svd.TruncatedSVD:
                                        n_components: 2
                            - gordo.machine.model.models.{model}:
                                kind: {model_kind}
                    """,
                        pydoc.locate(f"gordo.machine.model.models.{model}"),
                        model_kind,
                    ),
                    # Define pipeline memory with something other than None w/o metadata
                    (
                        f"""
                    sklearn.pipeline.Pipeline:
                        steps:
                        - sklearn.decomposition.pca.PCA:
                            n_components: 2
                            copy: true
                            whiten: false
                            svd_solver: auto
                            tol: 0.0
                            iterated_power: auto
                            random_state:
                        - sklearn.preprocessing._function_transformer.FunctionTransformer:
                                func: gordo.machine.model.transformer_funcs.general.multiply_by
                                kw_args:
                                    factor: 1
                        - sklearn.pipeline.FeatureUnion:
                            transformer_list:
                            - sklearn.decomposition.pca.PCA:
                                n_components: 3
                                copy: true
                                whiten: false
                                svd_solver: auto
                                tol: 0.0
                                iterated_power: auto
                                random_state:
                            - sklearn.pipeline.Pipeline:
                                steps:
                                - sklearn.preprocessing.data.MinMaxScaler:
                                    feature_range:
                                    - 0
                                    - 1
                                    copy: true
                                - sklearn.decomposition.truncated_svd.TruncatedSVD:
                                    n_components: 2
                                    algorithm: randomized
                                    n_iter: 5
                                    random_state:
                                    tol: 0.0
                                memory: /tmp
                            n_jobs: 1
                            transformer_weights:
                        - gordo.machine.model.models.{model}:
                            kind: {model_kind}
                    """,
                        pydoc.locate(f"gordo.machine.model.models.{model}"),
                        model_kind,
                    ),
                ]
                for template in templates:
                    yield template

    def test_pydoc_locate_class(self):
        self.factories = register_model_builder.factories
        for model in self.factories.keys():
            self.assertTrue(pydoc.locate(f"gordo.machine.model.models.{model}"))

    def test_from_definition(self):

        for raw_yaml, model, model_kind in self.setup_gen():
            self.assertTrue(model)
            logger.info(raw_yaml)
            config = yaml.load(raw_yaml)
            logger.debug("{}".format(config))

            config_clone = copy.deepcopy(config)  # To ensure no mutation occurs
            pipe = from_definition(config)

            # Test that the original config matches the one passed; no mutation
            self.assertEqual(config, config_clone)

            # Special tests that defining non-default argument holds for a
            # 'key:  ' is evaled to 'key=None'
            if "memory: /tmp" in raw_yaml:
                self.assertEqual(pipe.steps[2][1].transformer_list[1][1].memory, "/tmp")
            self._verify_pipe(pipe, model, model_kind)

    def _verify_pipe(self, pipe, model, model_kind):
        # We should now have a Pipeline.steps == 3
        self.assertTrue(len(pipe.steps), 3)

        # Note: All steps in a Pipeline or Transformer are tuples of (str, obj)
        # That is the reason for the semi-ugly indexing which follows.

        # STEP 1 TEST: Test expected PCA step
        step1 = pipe.steps[0][1]
        self.assertIsInstance(step1, PCA)
        self.assertEqual(step1.n_components, 2)

        # STEP 2 TEST: Test expected FunctionTransformer step
        step2 = pipe.steps[1][1]
        self.assertIsInstance(step2, FunctionTransformer)
        self.assertEqual(
            step2.func, gordo.machine.model.transformer_funcs.general.multiply_by
        )

        # STEP 3 TEST: Test expected FeatureUnion Step
        step3 = pipe.steps[2][1]
        self.assertIsInstance(step3, FeatureUnion)

        # First transformer of feature_transformers should be PCA(n_components=3)
        self.assertIsInstance(step3.transformer_list[0][1], PCA)
        self.assertEqual(step3.transformer_list[0][1].n_components, 3)

        # Third transformer in feature_transformers should be Pipeline
        sub_pipeline = step3.transformer_list[1][1]
        self.assertIsInstance(sub_pipeline, Pipeline)

        # First step in the sub pipeline is MinMaxScalar
        self.assertIsInstance(sub_pipeline.steps[0][1], MinMaxScaler)

        # Next step in the sub pipeline is TruncatedSVD w/ n_components=2
        self.assertIsInstance(sub_pipeline.steps[1][1], TruncatedSVD)
        self.assertEqual(sub_pipeline.steps[1][1].n_components, 2)

        # STEP 4 TEST:  Finally, the last step should be a KerasModel
        step4 = pipe.steps[3][1]
        self.assertIsInstance(step4, model)
        self.assertTrue(step4.kind, model_kind)



# -*- coding: utf-8 -*-
import collections
import copy
import re
import datetime

import pandas as pd
import dateutil.parser
import logging

from gordo.serializer import from_definition
from gordo.machine.dataset.sensor_tag import SensorTag


logger = logging.getLogger(__name__)


class BaseDescriptor:
    """
    Base descriptor class

    New object should override __set__(self, instance, value) method to check
    if 'value' meets required needs.
    """

    def __get__(self, instance, owner):
        return instance.__dict__[self.name]

    def __set_name__(self, owner, name):
        self.name = name

    def __set__(self, instance, value):
        raise NotImplementedError("Setting value not implemented for this Validator!")


class ValidDataset(BaseDescriptor):
    """
    Descriptor for attributes requiring type :class:`gordo.workflow.config_elements.Dataset`
    """

    def __set__(self, instance, value):

        # Avoid circular dependency imports
        from gordo.machine.dataset.base import GordoBaseDataset

        if not isinstance(value, GordoBaseDataset):
            raise TypeError(
                f"Expected value to be an instance of GordoBaseDataset, found {value}"
            )
        instance.__dict__[self.name] = value


class ValidDatasetKwargs(BaseDescriptor):
    """
    Descriptor for attributes requiring type :class:`gordo.workflow.config_elements.Dataset`
    """

    def _verify_resolution(self, resolution: str):
        """
        Verifies that a resolution string is supported in pandas
        """
        try:
            pd.tseries.frequencies.to_offset(resolution)
        except ValueError:
            raise ValueError(
                'Values for "resolution" must match pandas frequency terms: '
                "http://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html"
            )

    def __set__(self, instance, value):
        if not isinstance(value, dict):
            raise TypeError(f"Expected kwargs to be an instance of dict, found {value}")

        # Check that if 'resolution' is defined, it's one of supported pandas resampling frequencies
        if "resolution" in value:
            self._verify_resolution(value["resolution"])
        instance.__dict__[self.name] = value


class ValidModel(BaseDescriptor):
    """
    Descriptor for attributes requiring type Union[dict, str]
    """

    def __set__(self, instance, value):
        if getattr(instance, "_strict", True):
            try:
                from_definition(value)
            except Exception as e:
                raise ValueError(f"Pipeline from definition failed: {e}")
        instance.__dict__[self.name] = value


class ValidMetadata(BaseDescriptor):
    """
    Descriptor for attributes requiring type Optional[dict]
    """

    def __set__(self, instance, value):
        from gordo.machine.metadata import Metadata

        if value is not None and not any(
            isinstance(value, Obj) for Obj in (dict, Metadata)
        ):
            raise ValueError(f"Can either be None or an instance of dict or Metadata")
        instance.__dict__[self.name] = value


class ValidDataProvider(BaseDescriptor):
    """
    Descriptor for DataProvider
    """

    def __set__(self, instance, value):

        # Avoid circular dependency imports
        from gordo.machine.dataset.data_provider.base import GordoBaseDataProvider

        if not isinstance(value, GordoBaseDataProvider):
            raise TypeError(
                f"Expected value to be an instance of GordoBaseDataProvider, "
                f"found {value} "
            )
        instance.__dict__[self.name] = value


class ValidMachineRuntime(BaseDescriptor):
    """
    Descriptor for runtime dict in a machine object. Must be a valid runtime, but also
    must contain server.resources.limits/requests.memory/cpu to be valid.
    """

    def __set__(self, instance, value):
        if not isinstance(value, dict):
            raise ValueError(f"Runtime must be an instance of dict")
        value = self._verify_reporters(value)
        value = fix_runtime(value)
        instance.__dict__[self.name] = value

    @staticmethod
    def _verify_reporters(value: dict):
        """
        Verify the expected existence and structure of runtime.reporters
        """
        runtime = copy.deepcopy(value)
        if "reporters" not in runtime:
            runtime["reporters"] = list()
        else:
            assert isinstance(runtime["reporters"], list), "reporters should be a list"
        assert all(
            isinstance(rptr, dict) or isinstance(rptr, str)
            for rptr in runtime["reporters"]
        ), "All elements in 'reporters' should be a dict or str instances."
        return runtime


def fix_runtime(runtime_dict):
    """A valid runtime description must satisfy that any resource
    description must have that limit >= requests. This function will bump any limits
    that is too low."""
    runtime_dict = copy.deepcopy(runtime_dict)
    # We must also limit/request errors

    for key, val in runtime_dict.items():
        if isinstance(val, collections.Mapping):
            resource = val.get("resources")
            if resource:
                runtime_dict[key]["resources"] = fix_resource_limits(resource)
    return runtime_dict


def fix_resource_limits(resources: dict) -> dict:
    """
    Resource limitations must be higher or equal to resource requests, if they are
    both specified. This bumps any limits to the corresponding request if they are both
    set.

    Parameters
    ----------
    resources: dict
        Dictionary with possible requests/limits

    Examples
    --------
    >>> fix_resource_limits({"requests": {"cpu": 10}, "limits":{"cpu":9}})
    {'requests': {'cpu': 10}, 'limits': {'cpu': 10}}
    >>> fix_resource_limits({"requests": {"cpu": 10}})
    {'requests': {'cpu': 10}}


    Returns
    -------
    dict:
        A copy of `resource_dict` with the any limits bumped to the corresponding request if
        they are both set.
    """
    resources = copy.deepcopy(resources)
    requests = resources.get("requests", dict())
    limits = resources.get("limits", dict())
    request_memory = requests.get("memory")
    limits_memory = limits.get("memory")
    requests_cpu = requests.get("cpu")
    limits_cpu = limits.get("cpu")

    for r in [request_memory, limits_memory, requests_cpu, limits_cpu]:
        if r is not None and not isinstance(r, int):
            raise ValueError(
                f"Resource descriptions must be integers, and '{r}' is not."
            )
    if (
        limits_memory is not None
        and request_memory is not None
        and request_memory > limits_memory
    ):
        logger.warning(
            f"Memory limit {limits_memory} can not be smaller than memory "
            f"request {request_memory}, increasing memory limit to be equal"
            f" to request. "
        )
        limits["memory"] = request_memory
    if (
        limits_cpu is not None
        and requests_cpu is not None
        and requests_cpu > limits_cpu
    ):
        logger.warning(
            f"CPU limit {limits.get('cpu')} can not be smaller than cpu request"
            f" {requests.get('cpu')}, increasing cpu limit to be equal to request."
        )
        limits["cpu"] = requests_cpu
    return resources


class ValidDatetime(BaseDescriptor):
    """
    Descriptor for attributes requiring valid datetime.datetime attribute
    """

    def __set__(self, instance, value):
        datetime_value = None
        if isinstance(value, datetime.datetime):
            datetime_value = value
        elif isinstance(value, str):
            datetime_value = dateutil.parser.isoparse(value)
        else:
            raise ValueError(
                f"'{value}' is not a valid datetime.datetime object or string!"
            )

        if datetime_value.tzinfo is None:
            raise ValueError(f"Provide timezone to timestamp '{value}'")

        instance.__dict__[self.name] = datetime_value


class ValidTagList(BaseDescriptor):
    """
    Descriptor for attributes requiring a non-empty list of strings
    """

    def __set__(self, instance, value):
        if (
            len(value) == 0
            or not isinstance(value, list)
            or not any(isinstance(value[0], inst) for inst in (str, dict, SensorTag))
        ):
            raise ValueError(f"Requires setting a non-empty list of strings")
        instance.__dict__[self.name] = value


class ValidUrlString(BaseDescriptor):
    """
    Descriptor for use in objects which require valid URL values.
    Where 'valid URL values' is Gordo's version: alphanumeric with dashes.

    Use:

    .. code-block:: python

        class MySpecialClass:

            url_attribute = ValidUrlString()

            ...

        myspecialclass = MySpecialClass()

        myspecialclass.url_attribute = 'this-is-ok'
        myspecialclass.url_attribute = 'this will r@ise a ValueError'
    """

    def __set__(self, instance, value):
        if not self.valid_url_string(value):
            raise ValueError(
                f"'{value}' is not a valid Gordo url value. Only lower-case alphanumeric with dashes allowed.'"
            )
        if len(value) > 63:
            raise ValueError(
                f"'{value}' should be less than 63 chars, as required by Kubernetes/AKS DNS requirements."
            )
        instance.__dict__[self.name] = value

    @staticmethod
    def valid_url_string(string: str) -> bool:
        """
        What we (Gordo) deem to be a suitable URL is the same as kubernetes
        lowercase alphanumeric with dashes but not ending or starting with a dash

        Parameters
        ----------
            string: str - String to check

        Returns
        -------
            bool
        """
        return bool(
            re.match(
                r"^([a-z0-9]([-a-z0-9]*[a-z0-9])?(\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*)$",
                string,
            )
        )



# -*- coding: utf-8 -*-

import os
import io
import logging
import traceback
import timeit
import typing

import pandas as pd
from flask import Blueprint, current_app, g, send_file, make_response, jsonify, request
from flask_restplus import Resource, fields

from gordo import __version__, serializer
from gordo.server.rest_api import Api
from gordo.server import utils as server_utils
from gordo.machine.model import utils as model_utils
from gordo.machine.dataset.sensor_tag import SensorTag, normalize_sensor_tags
from gordo.server import model_io


logger = logging.getLogger(__name__)

base_blueprint = Blueprint("base_model_view", __name__)

api = Api(
    app=base_blueprint,
    title="Gordo Base Model View API Docs",
    version=__version__,
    description="Documentation for the Gordo ML Server",
    default_label="Gordo Endpoints",
)

# POST type declarations
API_MODEL_INPUT_POST = api.model(
    "Prediction - Multiple Samples", {"X": fields.List(fields.List(fields.Float))}
)
API_MODEL_OUTPUT_POST = api.model(
    "Prediction - Output from POST", {"output": fields.List(fields.List(fields.Float))}
)
_tags = {
    fields.String: fields.Float
}  # tags of single prediction record {'tag-name': tag-value}
_single_prediction_record = {
    "start": fields.DateTime,
    "end": fields.DateTime,
    "tags": fields.Nested(_tags),
    "total_abnormality": fields.Float,
}


class BaseModelView(Resource):
    """
    The base model view.
    """

    methods = ["POST"]

    y: pd.DataFrame = None
    X: pd.DataFrame = None

    @property
    def frequency(self):
        """
        The frequency the model was trained with in the dataset
        """
        return pd.tseries.frequencies.to_offset(g.metadata["dataset"]["resolution"])

    @property
    def tags(self) -> typing.List[SensorTag]:
        """
        The input tags for this model

        Returns
        -------
        typing.List[SensorTag]
        """
        return normalize_sensor_tags(
            g.metadata["dataset"]["tag_list"],
            asset=g.metadata["dataset"].get("asset"),
            default_asset=g.metadata["dataset"].get("default_asset"),
        )

    @property
    def target_tags(self) -> typing.List[SensorTag]:
        """
        The target tags for this model

        Returns
        -------
        typing.List[SensorTag]
        """
        if "target_tag_list" in g.metadata["dataset"]:
            return normalize_sensor_tags(
                g.metadata["dataset"]["target_tag_list"],
                asset=g.metadata["dataset"].get("asset"),
                default_asset=g.metadata["dataset"].get("default_asset"),
            )
        else:
            return []

    @api.response(200, "Success", API_MODEL_OUTPUT_POST)
    @api.expect(API_MODEL_INPUT_POST, validate=False)
    @api.doc(params={"X": "Nested or single list of sample(s) to predict"})
    @server_utils.model_required
    @server_utils.extract_X_y
    def post(self):
        """
        Process a POST request by using provided user data

        A typical response might look like this

        .. code-block:: python

            {
                'data': [
                    {
                        'end': ['2016-01-01T00:10:00+00:00'],
                        'model-output': [0.0005317790200933814,
                                         -0.0001525811239844188,
                                         0.0008310950361192226,
                                         0.0015755111817270517],
                        'original-input': [0.9135588550070414,
                                           0.3472517774179448,
                                           0.8994921857179736,
                                           0.11982773108991263],
                        'start': ['2016-01-01T00:00:00+00:00'],
                    },
                    ...
                ],

                'tags': [
                    {'asset': None, 'name': 'tag-0'},
                    {'asset': None, 'name': 'tag-1'},
                    {'asset': None, 'name': 'tag-2'},
                    {'asset': None, 'name': 'tag-3'}
                ],
                'time-seconds': '0.1937'
            }
        """
        context: typing.Dict[typing.Any, typing.Any] = dict()
        X = g.X
        process_request_start_time_s = timeit.default_timer()

        try:
            output = model_io.get_model_output(model=g.model, X=X)
        except ValueError as err:
            tb = traceback.format_exc()
            logger.error(
                f"Failed to predict or transform; error: {err} - \nTraceback: {tb}"
            )
            context["error"] = f"ValueError: {str(err)}"
            return make_response((jsonify(context), 400))

        # Model may only be a transformer, probably an AttributeError, but catch all to avoid logging other
        # exceptions twice if it happens.
        except Exception as exc:
            tb = traceback.format_exc()
            logger.error(
                f"Failed to predict or transform; error: {exc} - \nTraceback: {tb}"
            )
            context["error"] = "Something unexpected happened; check your input data"
            return make_response((jsonify(context), 400))

        else:
            get_model_output_time_s = timeit.default_timer()
            logger.debug(
                f"Calculating model output took "
                f"{get_model_output_time_s-process_request_start_time_s} s"
            )
            data = model_utils.make_base_dataframe(
                tags=self.tags,
                model_input=X.values if isinstance(X, pd.DataFrame) else X,
                model_output=output,
                target_tag_list=self.target_tags,
                index=X.index,
            )
            if request.args.get("format") == "parquet":
                return send_file(
                    io.BytesIO(server_utils.dataframe_into_parquet_bytes(data)),
                    mimetype="application/octet-stream",
                )
            else:
                context["data"] = server_utils.dataframe_to_dict(data)
                return make_response(
                    (jsonify(context), context.pop("status-code", 200))
                )


class MetaDataView(Resource):
    """
    Serve model / server metadata
    """

    @server_utils.metadata_required
    def get(self):
        """
        Get metadata about this endpoint, also serves as /healthcheck endpoint
        """
        model_collection_env_var = current_app.config["MODEL_COLLECTION_DIR_ENV_VAR"]
        return {
            "gordo-server-version": __version__,
            "metadata": g.metadata,
            "env": {model_collection_env_var: os.environ.get(model_collection_env_var)},
        }


class DownloadModel(Resource):
    """
    Download the trained model

    suitable for reloading via :func:`gordo.serializer.serializer.loads`
    """

    @api.doc(description="Download model, loadable via gordo.serializer.loads")
    @server_utils.model_required
    def get(self):
        """
        Responds with a serialized copy of the current model being served.

        Returns
        -------
        bytes
            Results from ``gordo.serializer.dumps()``
        """
        serialized_model = serializer.dumps(g.model)
        buff = io.BytesIO(serialized_model)
        return send_file(buff, attachment_filename="model.tar.gz")


class ModelListView(Resource):
    """
    List the current models capable of being served by the server
    """

    @api.doc(description="List the name of the models capable of being served.")
    def get(self, gordo_project: str):
        try:
            available_models = os.listdir(g.collection_dir)
        except FileNotFoundError:
            available_models = []
        finally:
            return jsonify({"models": available_models})


class RevisionListView(Resource):
    """
    List the available revisions the model can serve.
    """

    @api.doc(description="Available revisions of the project that can be served.")
    def get(self, gordo_project: str):
        try:
            available_revisions = os.listdir(os.path.join(g.collection_dir, ".."))
        except FileNotFoundError:
            logger.error(
                f"Attempted to list directories above {g.collection_dir} but failed with: {traceback.format_exc()}"
            )
            available_revisions = [g.current_revision]
        return jsonify(
            {"latest": g.current_revision, "available-revisions": available_revisions}
        )


class ExpectedModels(Resource):
    @api.doc(description="Models that the server expects to be able to serve.")
    def get(self, gordo_project: str):
        return jsonify({"expected-models": current_app.config["EXPECTED_MODELS"]})


api.add_resource(ModelListView, "/gordo/v0/<gordo_project>/models")
api.add_resource(ExpectedModels, "/gordo/v0/<gordo_project>/expected-models")
api.add_resource(BaseModelView, "/gordo/v0/<gordo_project>/<gordo_name>/prediction")
api.add_resource(
    MetaDataView,
    "/gordo/v0/<gordo_project>/<gordo_name>/metadata",
    "/gordo/v0/<gordo_project>/<gordo_name>/healthcheck",
)
api.add_resource(DownloadModel, "/gordo/v0/<gordo_project>/<gordo_name>/download-model")
api.add_resource(RevisionListView, "/gordo/v0/<gordo_project>/revisions")



import logging

import adal
import pytest
from unittest.mock import patch, call

from gordo.machine.dataset.data_provider.azure_utils import get_datalake_token


def test_get_datalake_token_wrong_args():
    with pytest.raises(ValueError):
        get_datalake_token()


@patch("azure.datalake.store.lib.auth")
def test_get_datalake_interactive(mock_stuff):
    get_datalake_token(interactive=True)
    assert mock_stuff.call_args_list == [call()]


@patch("azure.datalake.store.lib.auth")
def test_get_datalake_non_interactive(mock_stuff):
    get_datalake_token(
        interactive=False, dl_service_auth_str="TENTANT_UNKNOWN:BOGUS:PASSWORD"
    )
    assert mock_stuff.call_args_list == [
        call(
            client_id="BOGUS",
            client_secret="PASSWORD",
            resource="https://datalake.azure.net/",
            tenant_id="TENTANT_UNKNOWN",
        )
    ]


def test_get_data_serviceauth_fail(caplog):
    with pytest.raises(adal.adal_error.AdalError), caplog.at_level(logging.CRITICAL):
        get_datalake_token(dl_service_auth_str="TENTANT_UNKNOWN:BOGUS:PASSWORD")



# -*- coding: utf-8 -*-

import unittest

import pytest
from tensorflow.keras import optimizers

from gordo.machine.model.factories.lstm_autoencoder import (
    lstm_model,
    lstm_symmetric,
    lstm_hourglass,
)


class LSTMAutoEncoderTestCase(unittest.TestCase):
    def test_lstm_defaults(self):
        """
        Test that models with all defaults are created without failure and are equal
        """
        # Create models with default parameters
        base = lstm_model(4)
        symmetric = lstm_symmetric(4)
        hourglass = lstm_hourglass(4)

        # Ensure LSTM model layers are equal to base model layers
        for i in range(len(base.layers)):
            # Rename layers so as to not fail on names, only configuration
            config = base.layers[i].get_config().update({"name": "test"})
            symmetric_config = symmetric.layers[i].get_config().update({"name": "test"})
            hourglass_config = hourglass.layers[i].get_config().update({"name": "test"})
            assert config == symmetric_config
            assert config == hourglass_config

    def test_lstm_symmetric_checks_dims(self):
        """
        Test that lstm_symmetric validates parameter requirements
        """

        # Raise an error if empty dimension and function layers defined.
        with self.assertRaises(ValueError):
            lstm_symmetric(4, dims=(), funcs=())

        # Ensure failure with default 3 dimension layers when 2 function layers passed.
        with self.assertRaises(ValueError):
            lstm_symmetric(4, funcs=("tanh", "tanh"))

    def test_lstm_hourglass_basic(self):
        """
        Test that lstm_hourglass implements the correct parameters
        """

        model = lstm_hourglass(
            n_features=3,
            func="tanh",
            out_func="relu",
            optimizer="SGD",
            optimizer_kwargs={"lr": 0.02, "momentum": 0.001},
            compile_kwargs={"loss": "mae"},
        )

        # Ensure that the input dimension to Keras model matches the number of features.
        self.assertEqual(model.layers[0].input_shape[2], 3)

        # Ensure that the dimension of each encoding layer matches the expected dimension.
        self.assertEqual(
            [model.layers[i].input_shape[2] for i in range(1, 4)], [3, 2, 2]
        )

        # Ensure that the dimension of each decoding layer (excluding last decoding layer)
        # matches the expected dimension.
        self.assertEqual([model.layers[i].input_shape[2] for i in range(4, 6)], [2, 2])

        # Ensure that the dimension of last decoding layer matches the expected dimension.
        self.assertEqual(model.layers[6].input_shape[1], 3)

        # Ensure activation functions in the encoding part (layers 0-2)
        # match expected activation functions
        self.assertEqual(
            [model.layers[i].activation.__name__ for i in range(0, 3)],
            ["tanh", "tanh", "tanh"],
        )

        # Ensure activation functions in the decoding part (layers 3-5)
        # match expected activation functions
        self.assertEqual(
            [model.layers[i].activation.__name__ for i in range(3, 6)],
            ["tanh", "tanh", "tanh"],
        )

        # Ensure activation function for the output layer matches expected activation function
        self.assertEqual(model.layers[6].activation.__name__, "relu")

        # Assert that the expected Keras optimizer is used
        self.assertEqual(model.optimizer.__class__, optimizers.SGD)

        # Assert that the correct loss function is used.
        self.assertEqual(model.loss, "mae")

    def test_lstm_hourglass_checks_enc_layers(self):
        """
        Test that lstm_hourglass validates parameter requirements
        """
        with self.assertRaises(ValueError):
            lstm_hourglass(3, encoding_layers=0)

    def test_lstm_hourglass_checks_compression_factor(self):
        """
        Test that lstm_hourglass validates parameter requirements
        """
        with self.assertRaises(ValueError):
            lstm_hourglass(3, compression_factor=2)
        with self.assertRaises(ValueError):
            lstm_hourglass(3, compression_factor=-1)


@pytest.mark.parametrize("n_features", (5,))
@pytest.mark.parametrize("n_features_out", (10, 5, 2, 1))
def test_lstm_symmetric_basic(n_features, n_features_out):
    """
    Tests that lstm_symmetric implements the correct parameters
    """
    model = lstm_symmetric(
        n_features=n_features,
        n_features_out=n_features_out,
        lookback_window=3,
        dims=(4, 3, 2, 1),
        funcs=("relu", "relu", "tanh", "tanh"),
        out_func="linear",
        optimizer="SGD",
        optimizer_kwargs={"lr": 0.01},
        loss="mse",
    )

    # Ensure that the input dimension to Keras model matches the number of features.
    assert model.layers[0].input_shape[2] == n_features

    # Ensure that the dimension of each encoding layer matches the expected dimension.
    assert [model.layers[i].input_shape[2] for i in range(1, 5)] == [4, 3, 2, 1]

    # Ensure that the dimension of each decoding layer (excluding last decoding layer)
    # matches the expected dimension.
    assert [model.layers[i].input_shape[2] for i in range(5, 8)] == [1, 2, 3]

    # Ensure that the dimension of last decoding layer matches the expected dimension.
    assert model.layers[8].input_shape[1] == 4

    # Ensure activation functions in the encoding part (layers 0-3)
    # match expected activation functions.
    assert [model.layers[i].activation.__name__ for i in range(0, 4)] == [
        "relu",
        "relu",
        "tanh",
        "tanh",
    ]

    # Ensure activation functions in the decoding part (layers 4-7)
    # match expected activation functions.
    assert [model.layers[i].activation.__name__ for i in range(4, 8)] == [
        "tanh",
        "tanh",
        "relu",
        "relu",
    ]

    # Ensure activation function for the output layer matches expected activation function.
    assert model.layers[8].activation.__name__ == "linear"

    # Assert that the expected Keras optimizer is used
    assert model.optimizer.__class__ == optimizers.SGD

    # Assert that the correct loss function is used.
    assert model.loss == "mse"



import logging
import time
import pkg_resources
import json
import os

from typing import Dict, Any

import click

from gordo import __version__
from gordo.workflow.config_elements.normalized_config import NormalizedConfig
from gordo.workflow.workflow_generator import workflow_generator as wg


logger = logging.getLogger(__name__)


PREFIX = "WORKFLOW_GENERATOR"


@click.group("workflow")
@click.pass_context
def workflow_cli(gordo_ctx):
    pass


@click.command("generate")
@click.option(
    "--machine-config",
    type=str,
    help="Machine configuration file",
    envvar=f"{PREFIX}_MACHINE_CONFIG",
    required=True,
)
@click.option("--workflow-template", type=str, help="Template to expand")
@click.option(
    "--owner-references",
    type=wg._valid_owner_ref,
    default=None,
    allow_from_autoenv=True,
    help="Kubernetes owner references to inject into all created resources. "
    "Should be a nonempty yaml/json list of owner-references, each owner-reference "
    "a dict containing at least the keys 'uid', 'name', 'kind', and 'apiVersion'",
    envvar=f"{PREFIX}_OWNER_REFERENCES",
)
@click.option(
    "--gordo-version",
    type=str,
    default=wg._docker_friendly_version(__version__),
    help="Version of gordo to use, if different than this one",
    envvar=f"{PREFIX}_GORDO_VERSION",
)
@click.option(
    "--project-name",
    type=str,
    help="Name of the project which own the workflow.",
    allow_from_autoenv=True,
    envvar=f"{PREFIX}_PROJECT_NAME",
    required=True,
)
@click.option(
    "--project-revision",
    type=str,
    default=int(time.time() * 1000),  # unix time milliseconds
    help="Revision of the project which own the workflow.",
    envvar=f"{PREFIX}_PROJECT_REVISION",
)
@click.option(
    "--output-file",
    type=str,
    required=False,
    help="Optional file to render to",
    envvar=f"{PREFIX}_OUTPUT_FILE",
)
@click.option(
    "--namespace",
    type=str,
    default="kubeflow",
    help="Which namespace to deploy services into",
    envvar=f"{PREFIX}_NAMESPACE",
)
@click.option(
    "--split-workflows",
    type=int,
    default=50,
    help="Split workflows containg more than this number of models into several "
    "workflows, where each workflow contains at most this nr of models. The "
    "workflows are outputted sequentially with '---' in between, which allows "
    "kubectl to apply them all at once.",
    envvar=f"{PREFIX}_SPLIT_WORKFLOWS",
)
@click.option(
    "--n-servers",
    type=int,
    default=None,
    help="Max number of ML Servers to use, defaults to N machines * 10",
    envvar=f"{PREFIX}_N_SERVERS",
)
@click.option(
    "--docker-repository",
    type=str,
    default="gordo",
    help="The docker repo to use for pulling component images from",
    envvar=f"{PREFIX}_DOCKER_REPOSITORY",
)
@click.option(
    "--docker-registry",
    type=str,
    default="auroradevacr.azurecr.io",  # TODO: Change to docker.io after migrating
    help="The docker registry to use for pulling component images from",
    envvar=f"{PREFIX}_DOCKER_REGISTRY",
)
@click.pass_context
def workflow_generator_cli(gordo_ctx, **ctx):
    """
    Machine Configuration to Argo Workflow
    """

    context: Dict[Any, Any] = ctx.copy()
    yaml_content = wg.get_dict_from_yaml(context["machine_config"])

    try:
        log_level = yaml_content["globals"]["runtime"]["log_level"]
    except KeyError:
        log_level = os.getenv("GORDO_LOG_LEVEL", gordo_ctx.obj["log_level"])

    logging.getLogger("gordo").setLevel(log_level.upper())
    context["log_level"] = log_level.upper()

    # Create normalized config
    config = NormalizedConfig(yaml_content, project_name=context["project_name"])

    context["max_server_replicas"] = (
        context.pop("n_servers") or len(config.machines) * 10
    )

    # We know these exist since we set them in the default globals
    builder_resources = config.globals["runtime"]["builder"]["resources"]
    context["model_builder_resources_requests_memory"] = builder_resources["requests"][
        "memory"
    ]
    context["model_builder_resources_requests_cpu"] = builder_resources["requests"][
        "cpu"
    ]
    context["model_builder_resources_limits_memory"] = builder_resources["limits"][
        "memory"
    ]
    context["model_builder_resources_limits_cpu"] = builder_resources["limits"]["cpu"]

    context["server_resources"] = config.globals["runtime"]["server"]["resources"]

    # These are also set in the default globals, and guaranteed to exist
    client_resources = config.globals["runtime"]["client"]["resources"]
    context["client_resources_requests_memory"] = client_resources["requests"]["memory"]
    context["client_resources_requests_cpu"] = client_resources["requests"]["cpu"]
    context["client_resources_limits_memory"] = client_resources["limits"]["memory"]
    context["client_resources_limits_cpu"] = client_resources["limits"]["cpu"]

    context["client_max_instances"] = config.globals["runtime"]["client"][
        "max_instances"
    ]

    influx_resources = config.globals["runtime"]["influx"]["resources"]
    context["influx_resources_requests_memory"] = influx_resources["requests"]["memory"]
    context["influx_resources_requests_cpu"] = influx_resources["requests"]["cpu"]
    context["influx_resources_limits_memory"] = influx_resources["limits"]["memory"]
    context["influx_resources_limits_cpu"] = influx_resources["limits"]["cpu"]

    nr_of_models_with_clients = len(
        [
            machine
            for machine in config.machines
            if machine.runtime.get("influx", {}).get("enable", True)
        ]
    )
    context["client_total_instances"] = nr_of_models_with_clients

    # Should we start up influx/grafana at all, i.e. is there at least one request
    # for it?"
    enable_influx = nr_of_models_with_clients > 0
    context["enable_influx"] = enable_influx

    context["postgres_host"] = f"gordo-postgres-{config.project_name}"

    # If enabling influx, we setup a postgres reporter to send metadata
    # to allowing querying about the machine from grafana
    if enable_influx:
        pg_reporter = {
            "gordo.reporters.postgres.PostgresReporter": {
                "host": context["postgres_host"]
            }
        }
        for machine in config.machines:
            machine.runtime["reporters"].append(pg_reporter)

    # Determine if MlFlowReporter should be enabled per machine
    for machine in config.machines:
        try:
            enabled = machine.runtime["builder"]["remote_logging"]["enable"]
        except KeyError:
            continue
        else:
            if enabled:
                machine.runtime["reporters"].append(
                    "gordo.reporters.mlflow.MlFlowReporter"
                )

    context["machines"] = config.machines

    # Context requiring pre-processing
    context["target_names"] = [machine.name for machine in config.machines]

    # Json dump owner_references, if not None, otherwise pop it out of the context
    if context["owner_references"]:
        context["owner_references"] = json.dumps(context["owner_references"])
    else:
        context.pop("owner_references")

    if context["workflow_template"]:
        template = wg.load_workflow_template(context["workflow_template"])
    else:
        workflow_template = pkg_resources.resource_filename(
            "gordo.workflow.workflow_generator.resources", "argo-workflow.yml.template"
        )
        template = wg.load_workflow_template(workflow_template)

    # Clear output file
    if context["output_file"]:
        open(context["output_file"], "w").close()  # type: ignore
    for i in range(0, len(config.machines), context["split_workflows"]):  # type: ignore
        logger.info(
            f"Generating workflow for machines {i} to {i + context['split_workflows']}"
        )
        context["machines"] = config.machines[i : i + context["split_workflows"]]

        if context["output_file"]:
            s = template.stream(**context)
            with open(context["output_file"], "a") as f:  # type: ignore
                if i != 0:
                    f.write("\n---\n")
                s.dump(f)
        else:
            output = template.render(**context)
            if i != 0:
                print("\n---\n")
            print(output)


@click.command("unique-tags")
@click.option(
    "--machine-config", type=str, required=True, help="Machine configuration file"
)
@click.option(
    "--output-file-tag-list",
    type=str,
    required=False,
    help="Optional file to dump list of unique tags",
)
def unique_tag_list_cli(machine_config: str, output_file_tag_list: str):

    yaml_content = wg.get_dict_from_yaml(machine_config)

    machines = NormalizedConfig(yaml_content, project_name="test-proj-name").machines

    tag_list = set(tag for machine in machines for tag in machine.dataset.tag_list)

    if output_file_tag_list:
        with open(output_file_tag_list, "w") as output_file:
            for tag in tag_list:
                output_file.write(f"{tag.name}\n")
    else:
        for tag in tag_list:
            print(tag.name)


workflow_cli.add_command(workflow_generator_cli)
workflow_cli.add_command(unique_tag_list_cli)

if __name__ == "__main__":
    workflow_cli()



import os
import pytest
import logging
import re

from click.testing import CliRunner
from unittest import mock
import mlflow


from gordo import cli
from gordo.cli.cli import expand_model
from gordo.serializer import serializer
from gordo.machine import Machine
from tests.utils import temp_env_vars

import json

DATA_CONFIG = {
    "type": "RandomDataset",
    "train_start_date": "2015-01-01T00:00:00+00:00",
    "train_end_date": "2015-06-01T00:00:00+00:00",
    "tags": ["TRC1", "TRC2"],
    "target_tag_list": ["TRC1", "TRC2"],
}

DEFAULT_MODEL_CONFIG = {
    "gordo.machine.model.models.KerasAutoEncoder": {"kind": "feedforward_hourglass"}
}


MODEL_CONFIG = {"sklearn.decomposition.pca.PCA": {"svd_solver": "auto"}}
MODEL_CONFIG_WITH_PREDICT = {
    "gordo.machine.model.models.KerasAutoEncoder": {"kind": "feedforward_hourglass"}
}

logger = logging.getLogger(__name__)


@pytest.fixture
def machine():
    return Machine(
        name="test-model",
        model=MODEL_CONFIG,
        dataset=DATA_CONFIG,
        project_name="project-name",
    )


@pytest.fixture
def runner(tmpdir):
    mlflow.set_tracking_uri(f"file:{tmpdir}")
    yield CliRunner()


def test_build_env_args(runner, tmpdir, machine):
    """
    Instead of passing OUTPUT_DIR directly to CLI, should be able to
    read environment variables
    """
    logger.info(f"MODEL_CONFIG={json.dumps(MODEL_CONFIG)}")

    with temp_env_vars(MACHINE=json.dumps(machine.to_dict()), OUTPUT_DIR=str(tmpdir)):
        result = runner.invoke(cli.gordo, ["build"])

    assert result.exit_code == 0, f"Command failed: {result}, {result.exception}"
    assert (
        len(os.listdir(tmpdir)) > 1
    ), "Building was supposed to create at least two files (model and metadata) in OUTPUT_DIR, but it did not!"


def test_build_use_registry(runner, tmpdir, machine):
    """
    Using a registry causes the second build of a model to copy the first to the
    new location.
    """

    output_dir_1 = os.path.join(tmpdir, "dir1")
    output_dir_2 = os.path.join(tmpdir, "dir2")

    with temp_env_vars(
        MACHINE=json.dumps(machine.to_dict()),
        OUTPUT_DIR=output_dir_1,
        MODEL_REGISTER_DIR=os.path.join(tmpdir, "reg"),
    ):
        result1 = runner.invoke(cli.gordo, ["build"])

    assert result1.exit_code == 0, f"Command failed: {result1}"
    # OUTPUT_DIR is the only difference
    with temp_env_vars(
        MACHINE=json.dumps(machine.to_dict()),
        OUTPUT_DIR=output_dir_2,
        MODEL_REGISTER_DIR=os.path.join(tmpdir, "reg"),
    ):
        result2 = runner.invoke(cli.gordo, ["build"])
    assert result2.exit_code == 0, f"Command failed: {result2}"

    first_metadata = serializer.load_metadata(output_dir_1)
    second_metadata = serializer.load_metadata(output_dir_2)

    # The metadata contains the model build date, so if it got rebuilt these two
    # would be different
    assert (
        first_metadata["metadata"]["build_metadata"]["model"]["model_creation_date"]
        == second_metadata["metadata"]["build_metadata"]["model"]["model_creation_date"]
    )


def test_build_use_registry_bust_cache(runner, tmpdir, machine):
    """
    Even using a registry we get separate model-paths when we ask for models for
    different configurations.
    """

    output_dir_1 = os.path.join(tmpdir, "dir1")
    output_dir_2 = os.path.join(tmpdir, "dir2")

    with temp_env_vars(
        MACHINE=json.dumps(machine.to_dict()),
        OUTPUT_DIR=output_dir_1,
        MODEL_REGISTER_DIR=os.path.join(tmpdir, "reg"),
    ):
        result1 = runner.invoke(cli.gordo, ["build"])

    assert result1.exit_code == 0, f"Command failed: {result1}"

    # NOTE: Different train dates!
    machine.dataset = machine.dataset.from_dict(
        {
            "type": "RandomDataset",
            "train_start_date": "2019-01-01T00:00:00+00:00",
            "train_end_date": "2019-06-01T00:00:00+00:00",
            "tags": ["TRC1", "TRC2"],
        }
    )
    with temp_env_vars(
        MACHINE=json.dumps(machine.to_dict()),
        OUTPUT_DIR=output_dir_2,
        MODEL_REGISTER_DIR=os.path.join(tmpdir, "reg"),
    ):
        result2 = runner.invoke(cli.gordo, ["build"])
    assert result2.exit_code == 0, f"Command failed: {result2}"

    first_metadata = serializer.load_metadata(output_dir_1)
    second_metadata = serializer.load_metadata(output_dir_2)
    # The metadata contains the model build date, so if it got rebuilt these two
    # would be different
    assert (
        first_metadata["metadata"]["build_metadata"]["model"]["model_creation_date"]
        != second_metadata["metadata"]["build_metadata"]["model"]["model_creation_date"]
    )


def test_build_model_with_parameters(runner, tmpdir, machine):
    """
    It works to build a simple model with parameters set
    """
    machine._strict = False
    machine.model = """
    {
     "sklearn.decomposition.pca.PCA":
      {
        "svd_solver": "{{svd_solver}}",
        "n_components": {{n_components}}
      }
    }
    """

    svd_solver = "auto"
    n_components = 0.5

    logger.info(f"MODEL_CONFIG={json.dumps(machine.model)}")

    with temp_env_vars(MACHINE=json.dumps(machine.to_dict()), OUTPUT_DIR=str(tmpdir)):
        args = [
            "build",
            "--model-parameter",
            f"svd_solver,{svd_solver}",
            "--model-parameter",
            f"n_components,{n_components}",
        ]

        # Run it twice to ensure the model location in the location file
        # is only written once and not appended.
        for _ in range(2):

            result = runner.invoke(cli.gordo, args=args)

            assert (
                result.exit_code == 0
            ), f"Command failed: {result}, {result.exception}"
            assert (
                len(os.listdir(tmpdir)) > 1
            ), "Building was supposed to create at least two files (model and metadata) in OUTPUT_DIR, but it did not!"


def test_expand_model_default_works():
    assert expand_model(str(DEFAULT_MODEL_CONFIG), {}) == DEFAULT_MODEL_CONFIG


def test_expand_model_expand_works():
    model_params = {"kind": "hourglass", "num": 5}
    model_template = "{'gordo.machine.model.models.KerasAutoEncoder': {'kind': '{{kind}}', 'num': {{num}}}} "
    expected_model = {
        "gordo.machine.model.models.KerasAutoEncoder": {"kind": "hourglass", "num": 5}
    }
    assert expand_model(model_template, model_params) == expected_model


def test_expand_model_complains_on_missing_vars():
    model_params = {"kind": "hourglass"}
    model_template = "{'gordo.machine.model.models.KerasAutoEncoder': {'kind': '{{kind}}', 'num': {{num}}}} "
    with pytest.raises(ValueError):
        expand_model(model_template, model_params)


@pytest.mark.parametrize(
    "exception,exit_code",  # ArithmeticError is not in the mapping of exceptions to
    # exit codes, so it should default to 1
    [(FileNotFoundError, 30), (Exception, 1), (ArithmeticError, 1)],
)
def test_build_exit_code(exception, exit_code, runner, tmpdir, machine):
    """
    Test that cli build exists with different exit codes for different errors.
    """
    machine.model = MODEL_CONFIG_WITH_PREDICT
    machine.evaluation = {"cv_mode": "cross_val_only"}

    logger.info(f"MODEL_CONFIG={json.dumps(machine.model)}")
    with mock.patch(
        "gordo.cli.cli.ModelBuilder.build",
        mock.MagicMock(side_effect=exception, autospec=True, return_value=None),
    ):
        with temp_env_vars(
            MACHINE=json.dumps(machine.to_dict()), OUTPUT_DIR=str(tmpdir)
        ):
            result = runner.invoke(cli.gordo, ["build"])
            assert result.exit_code == exit_code


@pytest.mark.parametrize(
    "should_save_model, cv_mode",
    [(True, {"cv_mode": "full_build"}), (False, {"cv_mode": "cross_val_only"})],
)
def test_build_cv_mode(
    tmpdir, runner: CliRunner, should_save_model: bool, cv_mode: str, machine: Machine
):
    """
    Testing build with cv_mode set to full and cross_val_only. Checks that cv_scores are
    printed and model are only saved when using the default (full) value.
    """
    machine.model = MODEL_CONFIG_WITH_PREDICT
    machine.evaluation = cv_mode  # type: ignore

    logger.info(f"MODEL_CONFIG={json.dumps(machine.model)}")

    tmp_model_dir = os.path.join(tmpdir, "tmp")
    os.makedirs(tmp_model_dir, exist_ok=True)

    with temp_env_vars(MACHINE=json.dumps(machine.to_dict()), OUTPUT_DIR=tmp_model_dir):
        result = runner.invoke(cli.gordo, ["build", "--print-cv-scores"])
        assert result.exit_code == 0
        # Checks that the file is empty or not depending on the mode.
        if should_save_model:
            assert len(os.listdir(tmp_model_dir)) != 0
        else:
            assert len(os.listdir(tmp_model_dir)) == 0

        # Checks the output contains 'explained-variance_raw-scores'
        assert "r2-score" in result.output
        assert "mean-squared-error" in result.output
        assert "mean-absolute-error" in result.output
        assert "explained-variance-score" in result.output


@pytest.mark.parametrize(
    "should_save_model, cv_mode_1, cv_mode_2",
    [
        (True, {"cv_mode": "full_build"}, {"cv_mode": "cross_val_only"}),
        (False, {"cv_mode": "cross_val_only"}, {"cv_mode": "cross_val_only"}),
    ],
)
def test_build_cv_mode_cross_val_cache(
    tmpdir,
    should_save_model: bool,
    cv_mode_1: str,
    cv_mode_2: str,
    runner: CliRunner,
    machine: Machine,
):
    """
    Checks that cv_scores uses cache if ran after a full build. Loads the same model, and can
    print the cv_scores from them.
    """
    logger.info(f"MODEL_CONFIG={json.dumps(machine.model)}")

    machine.evaluation = cv_mode_1  # type: ignore
    with temp_env_vars(MACHINE=json.dumps(machine.to_dict()), OUTPUT_DIR=str(tmpdir)):
        runner.invoke(cli.gordo, ["build"])

    machine.evaluation = cv_mode_2  # type: ignore
    with temp_env_vars(MACHINE=json.dumps(machine.to_dict()), OUTPUT_DIR=str(tmpdir)):
        runner.invoke(cli.gordo, ["build"])

    if should_save_model:
        assert len(os.listdir(tmpdir)) > 0
    else:
        assert len(os.listdir(tmpdir)) == 0


def test_build_cv_mode_build_only(tmpdir, runner: CliRunner, machine: Machine):
    """
    Testing build with cv_mode set to build_only. Checks that OUTPUT_DIR gets a model
    saved to it. It also checks that the metadata contains cv-duration-sec=None and
    cv-scores={}
    """

    logger.info(f"MODEL_CONFIG={json.dumps(machine.model)}")
    machine.evaluation = {"cv_mode": "build_only"}

    with temp_env_vars(MACHINE=json.dumps(machine.to_dict()), OUTPUT_DIR=str(tmpdir)):

        metadata_file = f"{os.path.join(tmpdir, 'metadata.json')}"
        runner.invoke(cli.gordo, ["build"])

        # A model has been saved
        assert len(os.listdir(tmpdir)) != 0
        with open(metadata_file) as f:
            metadata_json = json.loads(f.read())
            assert (
                metadata_json["metadata"]["build_metadata"]["model"][
                    "cross_validation"
                ]["cv_duration_sec"]
                is None
            )
            assert (
                metadata_json["metadata"]["build_metadata"]["model"][
                    "cross_validation"
                ]["scores"]
                == {}
            )


@mock.patch("gordo.reporters.mlflow.get_spauth_kwargs")
@mock.patch("gordo.reporters.mlflow.get_workspace_kwargs")
@mock.patch("gordo.reporters.mlflow.get_mlflow_client")
def test_mlflow_reporter_set_cli_build(
    MockClient,
    mock_get_workspace_kwargs,
    mock_get_spauth_kwargs,
    monkeypatch,
    runner,
    tmpdir,
    machine,
):
    """
    Tests disabling MlFlow logging in cli, and missing env var when enabled
    """

    mlflow.set_tracking_uri(f"file:{tmpdir}")
    machine.runtime = dict(
        reporters=[{"gordo.reporters.mlflow.MlFlowReporter": dict()}]
    )

    with temp_env_vars(MACHINE=json.dumps(machine.to_dict()), OUTPUT_DIR=str(tmpdir)):
        # Logging enabled, without env vars set:
        # Raise error
        result = runner.invoke(cli.gordo, ["build"])
        result.exit_code != 0

        # Logging enabled, with env vars set:
        # Build success, remote logging executed
        with monkeypatch.context() as m:
            m.setenv("DL_SERVICE_AUTH_STR", "test:test:test")
            m.setenv("AZUREML_WORKSPACE_STR", "test:test:test")

            result = runner.invoke(cli.gordo, ["build"])
            result.exit_code == 1
            assert MockClient.called
            assert mock_get_workspace_kwargs.called
            assert mock_get_spauth_kwargs.called

        # Reset call counts
        for m in [MockClient, mock_get_workspace_kwargs, mock_get_spauth_kwargs]:
            m.reset_mock()

    # Logging not enabled:
    # Build success, remote logging not executed
    machine.runtime = dict(builder=dict(remote_logging=dict(enable=False)))
    with temp_env_vars(MACHINE=json.dumps(machine.to_dict()), OUTPUT_DIR=str(tmpdir)):
        result = runner.invoke(cli.gordo, ["build"])
        result.exit_code == 0
        assert not MockClient.called
        assert not mock_get_workspace_kwargs.called
        assert not mock_get_spauth_kwargs.called


@pytest.mark.parametrize(
    "arg,value,exception_expected",
    [
        # Valid values
        ("--host", "0.0.0.0", False),
        ("--host", "127.0.0.0", False),
        ("--port", 5555, False),
        ("--workers", 1, False),
        ("--log-level", "info", False),
        ("--log-level", "debug", False),
        # Invalid values
        ("--host", "0.0.0", True),
        ("--port", 0, True),
        ("--port", 70000, True),
        ("--workers", -1, True),
        ("--log-level", "badlevel", True),
    ],
)
def test_gunicorn_execution_hosts(runner, arg, value, exception_expected):
    """
    Test the validation of input parameters to the `run_server` function via the gordo cli
    """

    with mock.patch(
        "gordo.server.server.run_server",
        mock.MagicMock(return_value=None, autospec=True),
    ) as m:
        result = runner.invoke(cli.gordo, ["run-server", arg, value])

        assert (
            (result.exit_code != 0) if exception_expected else (result.exit_code == 0)
        )
        assert m.called_once_with(value)


def test_log_level_cli():
    """
    Test that the --log-level option in the CLI sets the correct log-level in the genreated config file.
    """

    runner = CliRunner()
    args = [
        "--log-level",
        "test_log_level",
        "workflow",
        "generate",
        "--machine-config",
        "examples/config.yaml",
        "--project-name",
        "test",
    ]

    result = runner.invoke(cli.gordo, args)

    # Find the value on the next line after the key GORDO_LOG_LEVEL
    gordo_log_levels = re.findall(
        r"(?<=GORDO_LOG_LEVEL\r|GORDO_LOG_LEVEL\n)[^\r\n]+", result.stdout
    )

    # Assert all the values to the GORDO_LOG_LEVEL key contains the correct log-level
    assert all(["TEST_LOG_LEVEL" in value for value in gordo_log_levels])



import abc
import pandas as pd
from typing import Optional
from datetime import timedelta
from sklearn.base import BaseEstimator

from gordo.machine.model.base import GordoBase


class AnomalyDetectorBase(BaseEstimator, GordoBase, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def anomaly(
        self, X: pd.DataFrame, y: pd.DataFrame, frequency: Optional[timedelta] = None
    ) -> pd.DataFrame:
        """
        Take X, y and optionally frequency; returning a dataframe containing
        anomaly score(s)
        """
        ...



# -*- coding: utf-8 -*-

import os
import typing
import sys
import json
from datetime import datetime
from pprint import pprint

import click
import yaml
import pandas as pd  # noqa
from requests import Session

from gordo.client import Client
from gordo import serializer
from gordo.machine.dataset.data_provider import providers
from gordo.cli.custom_types import IsoFormatDateTime, DataProviderParam, key_value_par
from gordo.client.forwarders import ForwardPredictionsIntoInflux


@click.group("client")
@click.option("--project", help="The project to target")
@click.option("--host", help="The host the server is running on", default="localhost")
@click.option("--port", help="Port the server is running on", default=443)
@click.option("--scheme", help="tcp/http/https", default="https")
@click.option("--batch-size", help="How many samples to send", default=100000)
@click.option("--parallelism", help="Maximum asynchronous jobs to run", default=10)
@click.option(
    "--metadata",
    type=key_value_par,
    multiple=True,
    default=(),
    help="Key-Value pair to be entered as metadata labels, may use this option multiple times. "
    "to be separated by a comma. ie: --metadata key,val --metadata some key,some value",
)
@click.option(
    "--session-config",
    type=yaml.safe_load,
    default="{}",
    help="Config json/yaml to set on the requests.Session object. Useful when needing to supply"
    "authentication parameters such as header keys. ie. --session-config {'headers': {'API-KEY': 'foo-bar'}}",
)
@click.pass_context
def client(ctx: click.Context, *args, **kwargs):
    """
    Entry sub-command for client related activities
    """
    # Setup the session with any attributes set by the user
    session_config = kwargs.pop("session_config", None)
    if session_config:
        session = Session()
        for key, value in session_config.items():
            setattr(session, key, value)
        kwargs["session"] = session

    ctx.obj = {"args": args, "kwargs": kwargs}


@click.command("predict")
@click.argument("start", type=IsoFormatDateTime())
@click.argument("end", type=IsoFormatDateTime())
@click.option(
    "--target",
    help="A list of machines to target. If not provided then target all machines in the"
    " project",
    default=[],
    multiple=True,
)
@click.option(
    "--data-provider",
    type=DataProviderParam(),
    envvar="DATA_PROVIDER",
    help="DataProvider dict encoded as json. Must contain a 'type' key with the name of"
    " a DataProvider as value.",
)
@click.option(
    "--output-dir",
    type=click.Path(exists=True),
    help="Save output prediction dataframes in a directory",
)
@click.option(
    "--influx-uri",
    help="Format: <username>:<password>@<host>:<port>/<optional-path>/<db_name>",
)
@click.option("--influx-api-key", help="Key to provide to the destination influx")
@click.option(
    "--influx-recreate-db",
    help="Recreate the desintation DB before writing",
    is_flag=True,
    default=False,
)
@click.option(
    "--forward-resampled-sensors",
    help="forward the resampled sensor values",
    is_flag=True,
    default=False,
)
@click.option(
    "--n-retries",
    help="Time client should retry failed predictions",
    type=int,
    default=5,
)
@click.option(
    "--parquet/--no-parquet",
    help="Use parquet serialization when sending and receiving data from server",
    default=True,
)
@click.pass_context
def predict(
    ctx: click.Context,
    start: datetime,
    end: datetime,
    target: typing.List[str],
    data_provider: providers.GordoBaseDataProvider,
    output_dir: str,
    influx_uri: str,
    influx_api_key: str,
    influx_recreate_db: bool,
    forward_resampled_sensors: bool,
    n_retries: int,
    parquet: bool,
):
    """
    Run some predictions against the target
    """
    ctx.obj["kwargs"].update(
        {
            "data_provider": data_provider,
            "forward_resampled_sensors": forward_resampled_sensors,
            "n_retries": n_retries,
            "use_parquet": parquet,
        }
    )

    client = Client(*ctx.obj["args"], **ctx.obj["kwargs"])

    if influx_uri is not None:
        client.prediction_forwarder = ForwardPredictionsIntoInflux(  # type: ignore
            destination_influx_uri=influx_uri,
            destination_influx_api_key=influx_api_key,
            destination_influx_recreate=influx_recreate_db,
            n_retries=n_retries,
        )

    # Fire off getting predictions
    predictions = client.predict(
        start, end, targets=target,
    )  # type: typing.Iterable[typing.Tuple[str, pd.DataFrame, typing.List[str]]]

    # Loop over all error messages for each result and log them
    click.secho(f"\n{'-' * 20} Summary of failed predictions (if any) {'-' * 20}")
    exit_code = 0
    for (_name, _df, error_messages) in predictions:
        for err_msg in error_messages:
            # Any error message indicates we encountered at least one error
            exit_code = 1
            click.secho(err_msg, fg="red")

    # Shall we write the predictions out?
    if output_dir is not None:
        for (name, prediction_df, _err_msgs) in predictions:
            prediction_df.to_csv(
                os.path.join(output_dir, f"{name}.csv.gz"), compression="gzip"
            )
    sys.exit(exit_code)


@click.command("metadata")
@click.option(
    "--output-file",
    type=click.File(mode="w"),
    help="Optional output file to save metadata",
)
@click.option(
    "--target",
    help="A list of machines to target. If not provided then target all machines in the"
    " project",
    default=[],
    multiple=True,
)
@click.pass_context
def metadata(
    ctx: click.Context,
    output_file: typing.Optional[typing.IO[str]],
    target: typing.List[str],
):
    """
    Get metadata from a given endpoint
    """
    client = Client(*ctx.obj["args"], **ctx.obj["kwargs"])
    metadata = {
        k: v.to_dict() for k, v in client.get_metadata(targets=target).items()  # type: ignore
    }
    if output_file:
        json.dump(metadata, output_file)
        click.secho(f"Saved metadata json to file: '{output_file}'")
    else:
        pprint(metadata)
    return metadata


@click.command("download-model")
@click.argument("output-dir", type=click.Path(exists=True))
@click.option(
    "--target",
    help="A list of machines to target. If not provided then target all machines in the"
    " project",
    default=[],
    multiple=True,
)
@click.pass_context
def download_model(ctx: click.Context, output_dir: str, target: typing.List[str]):
    """
    Download the actual model from the target and write to an output directory
    """
    client = Client(*ctx.obj["args"], **ctx.obj["kwargs"])
    models = client.download_model(targets=target)

    # Iterate over mapping of models and save into their own sub dirs of the output_dir
    for model_name, model in models.items():
        model_out_dir = os.path.join(output_dir, model_name)
        os.mkdir(model_out_dir)
        click.secho(
            f"Writing model '{model_name}' to directory: '{model_out_dir}'...", nl=False
        )
        serializer.dump(model, model_out_dir)
        click.secho(f"done")

    click.secho(f"Wrote all models to directory: {output_dir}", fg="green")


client.add_command(predict)
client.add_command(metadata)
client.add_command(download_model)



import logging
import dateutil.parser

import pytest

from gordo.machine.dataset.data_provider.providers import InfluxDataProvider
from gordo.client.utils import influx_client_from_uri
from gordo.machine.dataset.dataset import _get_dataset


logger = logging.getLogger(__name__)


def test_read_single_sensor_empty_data_time_range_indexerror(
    influxdb, influxdb_uri, sensors_str, caplog
):
    """
    Asserts that an IndexError is raised because the dates requested are outside the existing time period
    """
    train_start_date = dateutil.parser.isoparse("2017-01-01T09:11:00+00:00")
    train_end_date = dateutil.parser.isoparse("2017-01-01T10:30:00+00:00")

    ds = InfluxDataProvider(
        measurement="sensors",
        value_name="Value",
        client=influx_client_from_uri(uri=influxdb_uri, dataframe_client=True),
    )

    with caplog.at_level(logging.CRITICAL):
        with pytest.raises(IndexError):
            ds.read_single_sensor(
                train_start_date=train_start_date,
                train_end_date=train_end_date,
                tag=sensors_str[0],
                measurement="sensors",
            )


def test_read_single_sensor_empty_data_invalid_tag_name_valueerror(
    influxdb, influxdb_uri
):
    """
    Asserts that a ValueError is raised because the tag name inputted is invalid
    """
    train_start_date = dateutil.parser.isoparse("2016-01-01T09:11:00+00:00")
    train_end_date = dateutil.parser.isoparse("2016-01-01T10:30:00+00:00")

    ds = InfluxDataProvider(
        measurement="sensors",
        value_name="Value",
        client=influx_client_from_uri(uri=influxdb_uri, dataframe_client=True),
    )
    with pytest.raises(ValueError):
        ds.read_single_sensor(
            train_start_date=train_start_date,
            train_end_date=train_end_date,
            tag="tag-does-not-exist",
            measurement="sensors",
        )


def test__list_of_tags_from_influx_validate_tag_names(
    influxdb, influxdb_uri, sensors_str
):
    """
    Test expected tags in influx match the ones actually in influx.
    """
    ds = InfluxDataProvider(
        measurement="sensors",
        value_name="Value",
        client=influx_client_from_uri(uri=influxdb_uri, dataframe_client=True),
    )
    list_of_tags = ds._list_of_tags_from_influx()
    expected_tags = sensors_str
    tags = set(list_of_tags)
    assert set(expected_tags) == tags, (
        f"Expected tags = {expected_tags}" f"outputted {tags}"
    )


def test_get_list_of_tags(influxdb, influxdb_uri, sensors_str):
    ds = InfluxDataProvider(
        measurement="sensors",
        value_name="Value",
        client=influx_client_from_uri(uri=influxdb_uri, dataframe_client=True),
    )
    expected_tags = set(sensors_str)

    tags = set(ds.get_list_of_tags())
    assert expected_tags == tags

    # The cache does not screw stuff up
    tags = set(ds.get_list_of_tags())
    assert expected_tags == tags


def test_influx_dataset_attrs(influxdb, influxdb_uri, sensors):
    """
    Test expected attributes
    """
    train_start_date = dateutil.parser.isoparse("2016-01-01T09:11:00+00:00")
    train_end_date = dateutil.parser.isoparse("2016-01-01T10:30:00+00:00")
    tag_list = sensors
    config = {
        "type": "TimeSeriesDataset",
        "train_start_date": train_start_date,
        "train_end_date": train_end_date,
        "tag_list": tag_list,
    }
    config["data_provider"] = InfluxDataProvider(
        measurement="sensors",
        value_name="Value",
        client=influx_client_from_uri(uri=influxdb_uri, dataframe_client=True),
    )
    dataset = _get_dataset(config)
    assert hasattr(dataset, "get_metadata")

    metadata = dataset.get_metadata()
    assert isinstance(metadata, dict)


def test_influx_load_series_dry_run_raises(sensors):
    ds = InfluxDataProvider(measurement="sensors", value_name="Value", client=None)
    train_start_date = dateutil.parser.isoparse("2016-01-01T09:11:00+00:00")
    train_end_date = dateutil.parser.isoparse("2016-01-01T10:30:00+00:00")
    tag_list = sensors
    with pytest.raises(NotImplementedError):
        ds.load_series(
            train_start_date=train_start_date,
            train_end_date=train_end_date,
            tag_list=tag_list,
            dry_run=True,
        )
